"""Apply basic maths transformations to be applied to timeseries of physical values.

This module defines utilities used to create fine-sampled timeseries from
coarse sampled one:

    - :py:func:`piecewise_affine`
    - :py:func:`piecewise_constant`

Resampling to coarser resolution may be done as well, but the relevance may
be questioned VS a well-chosen aggregation.

.. seealso:: In case the series to be resampled must satisfy conservation laws

    :py:mod:`energy_analysis_toolbox.timeseries.power.conservative`

"""

import numpy as np
import pandas as pd


def piecewise_affine(
    timeseries: pd.Series | float,
    target_instants: pd.DatetimeIndex,
) -> pd.Series:
    """Return resampled timeseries assuming a piecewise affine function of time.

    Parameters
    ----------
    timeseries : pd.Series or float
        Series of values of a function of time, indexed using DateTimeIndex.
    target_instants : pd.DatetimeIndex
        Dates at which the series values are required, sorted in ascending
        order.

    Returns
    -------
    new_series : pd.Series
        Values of the function, interpolated at target times, indexed with
        ``target_instants``.


    .. warning::

        The returned values may not be relevant when some target times are
        required outside the convex span of the input samples: the
        corresponding border value is used for these target times.


    .. seealso::

        :py:func:`np.interp` on which the interpolation is based.


    """
    try:
        ref_time = target_instants[0]
    except IndexError:
        return pd.Series([], dtype=timeseries.dtype, index=target_instants.copy())
    target_offsets = (target_instants - ref_time).total_seconds()
    sample_dts = (timeseries.index - ref_time).total_seconds()
    new_values = np.interp(target_offsets, sample_dts, timeseries.values)
    new_series = pd.Series(
        new_values,
        index=target_instants.copy(),
        name=timeseries.name,
    )
    new_series.index.name = timeseries.index.name
    return new_series


def piecewise_constant(
    timeseries: pd.Series,
    target_instants: pd.DatetimeIndex,
    left_pad: float | None = None,
) -> pd.Series:
    """Return resampled timeseries assuming a piecewise constant function of time.

    Parameters
    ----------
    timeseries : pd.Series or float
        Series of values of a function of time, indexed using DateTimeIndex.
    target_instants : pd.DatetimeIndex
        A sequence of target timestamps, sorted in ascending order.
    left_pad : float or None, optional
        A value to be used for target instants which are located before the
        first instant in ``timeseries``. The default is |None| in which case, the
        value of the first instant is used.

    Returns
    -------
    new_series : pd.Series
        Values of the series interpolated at target times, indexed with
        ``target_instants``.


    .. seealso::

        :py:func:`np.digitize` on which the function is based.

    """
    try:
        ref_time = target_instants[0]
    except IndexError:
        return pd.Series([], dtype=timeseries.dtype, index=target_instants.copy())
    target_offsets = (target_instants - ref_time).total_seconds()
    sample_dts = (timeseries.index - ref_time).total_seconds()
    ix_select = np.digitize(target_offsets, sample_dts, right=False) - 1
    if left_pad is None:
        ix_select[ix_select < 0] = 0  # index 0
        new_values = timeseries.iloc[ix_select].to_numpy()
    else:
        new_values = timeseries.iloc[ix_select].to_numpy()
        new_values[ix_select < 0] = left_pad
    new_series = pd.Series(new_values, index=target_instants, name=timeseries.name)
    new_series.index.name = timeseries.index.name
    return new_series



"""Contains basic tools to extract features in timeseries."""

import numpy as np
import pandas as pd

from energy_analysis_toolbox import keywords as eatk
from energy_analysis_toolbox.errors import (
    EATEmptyDataError,
    EATInvalidTimestepDurationError,
    EATUndefinedTimestepError,
)


def intervals_over(
    series: pd.Series,
    low_tshd: float,
    *,
    return_positions: bool = False,
) -> pd.DataFrame:
    """Return the overconsumption limits when the series values are over ``low_tshd``.

    Parameters
    ----------
    series : pd.Series
         The series in which overconsumption of consecutive values over ``low_tshd``
         are searched.
    low_tshd : float
        Lower threshold on the values in the series for interval extractions :
        consecutive values of the series elements over **(strict)** this threshold
        are searched.
    return_positions : bool, default False
        If ``True``, return a second Dataframe with interval bounds as integer
        positions in the provided ``series`` instead of the labels. Default is
        ``False``.

    Returns
    -------
    overconsumption : pd.DataFrame
        The dataframe of overconsumption, with two columns: |eatk.start_f| and
        |eatk.end_f|. Each row contains :

        - in |eatk.start_f| the label of the first instant of an interval when
          the values in the series are > `low_tshd`.
        - in |eatk.end_f| the label of the first instant after this interval

        such that the row describes interval as ``[start, end[``.
    iloc_bounds : pd.Dataframe, optional
        Dataframe with same structure as ``overconsumption``, except that the values
        for the start and ends of overconsumption are integer positions ("ilocs") in
        the original series instead of labels.

    Notes
    -----
    The algorithm used in this function is the following.

    [1] First, locate the instants when the value of ``series`` changes either:

    - from a value ``=< low_tshd`` to value ``> low_tshd``
    - from a value ``> low_tshd`` to value ``=< low_tshd``

    These shifts are interval limits.

    [2] The limits of the timeseries are special cases. These bounds are
    considered as shifts if the value is over the threshold at these positions.

    Recalling the indexation is for overconsumption which are open on the right side :

    - an interval start is the index of a positive change in values
      (first index of value ``> low_tshd``),
    - an interval end is the index of a negative change (first index
      ``=< low_tshd`` value).

    [3] Accordingly, with [1] and [2], shifts with even indices are interval
    starts while those with odd ones are interval ends.

    [4] The dataframe of interval bounds is assembled and returned. The ilocs
    are returned as well in case ``return_positions`` is ``True``.

    Example
    -------
    >>> time_begin = pd.Timestamp("2018-07-06 05:00:00")
    >>> time_range = pd.date_range(
            start=time_begin,
            periods=8,
            inclusive='left',
            freq=pd.DateOffset(seconds=300))
    >>> series = pd.Series(np.array([0, 1, 1.5, 2., 2., 3., 1.5, 0.]), index=time_range)
    >>> intervals_over(series, 1.5)
                    start                 end
        0 2018-07-06 05:15:00 2018-07-06 05:30:00
    >>> intervals_over(series, 42)
        Empty DataFrame
        Columns: [start, end]
        Index: []

    """
    if series.empty:
        return pd.DataFrame([], columns=[eatk.start_f, eatk.end_f])
    # 1 diffs in a Bool series -> True if element differs from previous
    over_status_shift = (series > low_tshd).diff()
    # 2
    over_status_shift.iloc[0] = series.iloc[0] > low_tshd
    over_status_shift.iloc[-1] = over_status_shift.iloc[-1] or (
        series.iloc[-1] > low_tshd
    )
    # indexes of shifts
    actual_times = over_status_shift.index.where(over_status_shift).dropna()
    actual_inds = [series.index.get_loc(time) for time in actual_times]
    # 3 as [[start, end]]
    bound_indexes = [
        [actual_inds[2 * i], actual_inds[2 * i + 1]]
        for i in range(len(actual_inds) // 2)
    ]
    bound_labels = [
        [over_status_shift.index[s], over_status_shift.index[e]]
        for s, e in bound_indexes
    ]
    intervals = pd.DataFrame(bound_labels, columns=[eatk.start_f, eatk.end_f])
    # 4
    if return_positions:
        iloc_bounds = pd.DataFrame(bound_indexes, columns=[eatk.start_f, eatk.end_f])
        return intervals, iloc_bounds
    return intervals


def timestep_durations(
    timeseries: pd.Series,
    last_step: float | None = None,
) -> pd.Series:
    """Return the series of timestep durations of a timeseries.

    Parameters
    ----------
    timeseries : pd.Series
        A timeseries in chronological order.
    last_step : float, optional
        Duration of the last time-step in the series in (s).
        The default is |None| meaning that the same duration as the former-last
        one is used.

    Raises
    ------
    EATEmptyDataError :
        If the series is empty.
    EATUndefinedTimestepError :
        If the series contains only one element and ``last_step``
        is |None|.

    Returns
    -------
    pd.Series
        The series of durations in (s). The element at iloc i is the duration
        before the next index in the series.


    .. important::

        The input timeseries should contain at least two elements (one if the
        duration of the last time-step is provided).

    .. seealso::

        :py:func:`energy_analysis_toolbox.timeseries.extract_features.basics.index_to_timesteps`
        which works directly from the series index.

    """
    if timeseries.empty:
        err = "Interval durations cannot be inferred for empty series."
        raise EATEmptyDataError(err)
    min_timeseries_length = 2
    if timeseries.size < min_timeseries_length and last_step is None:
        err = (
            "The timeseries should contain at least 2 elements to infer a duration"
            "when last_step value is None."
        )
        raise EATUndefinedTimestepError(err)
    return pd.Series(
        index_to_timesteps(timeseries.index, last_step),
        index=timeseries.index,
    )


def index_to_timesteps(
    time_indexes: pd.DatetimeIndex,
    last_step: float | None = None,
) -> np.array:
    """Return the array of interval durations of a DatetimeIndex.

    Parameters
    ----------
    time_indexes : pd.DatetimeIndex
        A sequence of time-steps in chronological order.
    last_step : float, optional
        Duration of the last time-step in the series in (s).
        The default is |None| meaning that the same duration as the former-last
        one is used.

    Raises
    ------
    EATEmptyDataError :
        If ``time_indexes`` is empty.
    EATUndefinedTimestepError :
        If ``time_indexes``  contains only one element and ``last_step``
        is |None|.
    EATInvalidTimestepDurationError :
        If ``last_step < 0``.

    Returns
    -------
    np.array
        The series of durations in (s). The element i is the duration
        before the next index.


    .. important::

        The input ``time_indexes`` should contain at least two elements (one if the
        duration of the last time-step is provided).

    .. seealso::

        :py:func:`energy_analysis_toolbox.timeseries.extract_features.basics.timestep_durations`
        which works on timeseries by applying this function to its index.

    """
    if time_indexes.empty:
        err = "Interval durations cannot be inferred for empty time-sequences."
        raise EATEmptyDataError(err)
    min_time_indices_length = 2
    if time_indexes.size < min_time_indices_length and last_step is None:
        err = (
            "The series should contain at least 2 elements to infer a duration"
            "when last_step value is None."
        )
        raise EATUndefinedTimestepError(err)
    if last_step is not None and last_step < 0:
        err = f"Last step duration must be >=0. Received {last_step} s."
        raise EATInvalidTimestepDurationError(err)
    # better initialize and assign to avoid table extension in the next step
    durations = np.empty(time_indexes.size)
    durations[:-1] = (time_indexes[1:] - time_indexes[0:-1]).total_seconds()
    if last_step is None:
        durations[-1] = durations[-2]
    else:
        durations[-1] = last_step
    return durations



import numpy as np
import pandas as pd
import pytest

from ..errors import EATUndefinedTimestepError
from ..power import integrate_over, to_energy


def _constant_power():
    """Return a power series of constant 1W value across a DST"""
    instants = pd.date_range(
        start=pd.Timestamp("2023-10-29", tz="Europe/Paris"),
        end=pd.Timestamp("2023-10-31", tz="Europe/Paris"),
        freq="1h",
        inclusive="left",
    )
    power_series = pd.Series(np.ones_like(instants), index=instants)
    return power_series


# ==============================================================================
# Test to_energy
# ==============================================================================
def test_to_energy():
    """Check that constant 1W power across DST is correctly transformed to energy."""
    power_series = _constant_power()
    energy = to_energy(power_series)
    expected = power_series * 3600
    pd.testing.assert_series_equal(energy, expected)


def test_to_energy_failure():
    """Check that the function fails on series with only one element."""
    power_series = _constant_power().iloc[:1]
    with pytest.raises(EATUndefinedTimestepError):
        to_energy(power_series)


# ==============================================================================
# Test integrate_over
# ==============================================================================


def _intervals():
    intervals = pd.DataFrame.from_dict(
        {
            "start": [
                pd.Timestamp("2023-10-29 00:00", tz="Europe/Paris"),
                pd.Timestamp("2023-10-30 02:00", tz="UTC+02:00"),
            ],
            "end": [
                pd.Timestamp("2023-10-29 01:00", tz="Europe/Paris"),
                pd.Timestamp("2023-10-30 02:00", tz="UTC+01:00"),
            ],
        },
    )
    intervals.index = ["A", "B"]
    return intervals


def test_integrate_over_empty_intervals():
    """Check that empty overconsumption leads to empty results"""
    no_intervals = _intervals().iloc[:0, :]
    assert integrate_over(no_intervals, _constant_power()).empty


def test_integrate_over():
    """Check that integration works across and DST"""
    intervals = _intervals()
    expected = pd.Series([3600.0, 3600.0], index=["A", "B"])
    pd.testing.assert_series_equal(
        integrate_over(intervals, _constant_power()),
        expected,
    )


def test_integrate_over_not_in_index():
    """Check that the function works when the bounds are not indices of the series"""
    intervals = _intervals()
    intervals["start"] -= pd.Timedelta("10min")
    expected = pd.Series([3600.0, 3600.0], index=["A", "B"])
    pd.testing.assert_series_equal(
        integrate_over(intervals, _constant_power()),
        expected,
    )
    # same on the right. Go one step beyond on the right.
    intervals = _intervals()
    intervals["end"] += pd.Timedelta("10min")
    expected = pd.Series([7200.0, 7200.0], index=["A", "B"])
    pd.testing.assert_series_equal(
        integrate_over(intervals, _constant_power()),
        expected,
    )



"""Test the eat.timeseries.resample.index_transformation module"""

import numpy as np
import pandas as pd
import pytest

from ..timeseries.resample.index_transformation import (
    estimate_timestep,
    fill_data_holes,
    fill_missing_entries,
    index_to_freq,
    tz_convert_or_localize,
)


@pytest.mark.parametrize(
    "source_tz, target_tz",
    [
        ("UTC", "Europe/Paris"),
        ("Europe/Paris", "UTC"),
        (None, "Europe/Paris"),
        ("Europe/Paris", None),
        (None, None),
    ],
)
def test_tz_convert_localize(source_tz, target_tz):
    """Check that the function works as expected on a series and dataframe"""
    a_timeseries = pd.Series(
        np.arange(10),
        index=pd.date_range("2020-01-01", periods=10, freq="1h", tz=source_tz),
        name="example_series",
    )
    a_dataframe = pd.DataFrame.from_dict(
        {
            "a": a_timeseries,
            "b": a_timeseries * 2,
        },
    )
    if a_timeseries.index.tz is None:
        expected_series = a_timeseries.tz_localize(target_tz)
        expected_df = a_dataframe.tz_localize(target_tz)
    else:
        expected_series = a_timeseries.tz_convert(target_tz)
        expected_df = a_dataframe.tz_convert(target_tz)
    pd.testing.assert_series_equal(
        tz_convert_or_localize(a_timeseries, tz=target_tz), expected_series,
    )
    pd.testing.assert_frame_equal(
        tz_convert_or_localize(a_dataframe, tz=target_tz), expected_df,
    )


def test_index_to_freq():
    """Test the index to freq nominal use"""
    index = pd.DatetimeIndex(
        [
            pd.Timestamp("2022-06-15 12:03"),
            pd.Timestamp("2022-06-15 12:05"),
            pd.Timestamp("2022-06-15 12:08"),
            pd.Timestamp("2022-06-15 12:13"),
            pd.Timestamp("2022-06-15 12:19"),
            pd.Timestamp("2022-06-15 12:20"),
        ],
    )
    freq = "1min"
    expected_index = pd.DatetimeIndex(
        pd.date_range(
            start=pd.Timestamp("2022-06-15 12:03"),
            end=pd.Timestamp("2022-06-15 12:20"),
            freq=freq,
        ),
    )
    returned_index = index_to_freq(index, freq=freq)
    pd.testing.assert_index_equal(expected_index, returned_index)


def test_estimate_timestep_regular_index():
    """Test that the expected timestep is the same for all method when the
    index is regular
    """
    freq = "1min"
    timestep = pd.Timedelta(freq).total_seconds()
    index = pd.DatetimeIndex(
        pd.date_range(
            start=pd.Timestamp("2022-06-15 12:03"),
            end=pd.Timestamp("2022-06-15 12:20"),
            freq=freq,
        ),
    )
    assert estimate_timestep(index, "median") == timestep
    assert estimate_timestep(index, "mean") == timestep
    assert estimate_timestep(index, "mode") == timestep
    # assert estimate_timestep(index, "KDE") == timestep  # KDE doesn't work with regularly spaced data


def test_estimate_timestep_missing_entry():
    """Test the expected timestep when one entry is missing"""
    freq = "1min"
    timestep = pd.Timedelta(freq).total_seconds()
    index = pd.DatetimeIndex(
        pd.date_range(
            start=pd.Timestamp("2022-06-15 12:03"),
            end=pd.Timestamp("2022-06-15 12:20"),
            freq=freq,
        ),
    )
    index = index.drop(pd.Timestamp("2022-06-15 12:15"))
    assert estimate_timestep(index, "median") == timestep
    assert estimate_timestep(index, "kde") == timestep
    assert estimate_timestep(index, "mode") == timestep
    # assert estimate_timestep(index, "mean") == timestep # this test doesn't work with mean


def test_estimate_timestep_two_frequency():
    """Test that if 2 frequency are present, then only one is returned.
    From tinkering with this test, it looks like it is the last frequency that is returned.
    """
    freq1 = "1min"
    freq2 = "2min"
    timestamp_1 = list(
        pd.date_range(start=pd.Timestamp("2022-06-15 12:03"), periods=10, freq=freq1),
    )
    timestamp_2 = list(
        pd.date_range(start=timestamp_1[-1], periods=10, freq=freq2, inclusive="right"),
    )
    index = pd.DatetimeIndex(timestamp_1 + timestamp_2)
    timestep = pd.Timedelta(freq2).total_seconds()
    assert estimate_timestep(index, "median") == timestep
    assert estimate_timestep(index, "kde") == timestep
    assert estimate_timestep(index, "mode") == timestep


def test_estimate_timestep_random_frequency():
    """Test that if the timestep are truly random, the return value still make sense."""
    timestep = 60  # seconds
    size = 1000
    timesteps = np.random.normal(timestep, 2, size=size)
    deltas = np.cumsum(timesteps)
    start = pd.Timestamp("2022-06-15 12:03")
    index = pd.DatetimeIndex([start + pd.Timedelta(seconds=s) for s in deltas])
    atol = 0.1
    rtol = 0.1
    assert np.isclose(
        estimate_timestep(index, "median"), timestep, rtol=rtol, atol=atol,
    )
    assert np.isclose(estimate_timestep(index, "kde"), timestep, rtol=rtol, atol=atol)
    assert np.isclose(
        estimate_timestep(index, "mode"), timestep, rtol=rtol, atol=atol,
    )  # I'm quite surprised that this works !
    assert np.isclose(estimate_timestep(index, "mean"), timestep, rtol=rtol, atol=atol)


def test_fill_missing_entry_series():
    """Test the fill missing entry function"""
    freq = "1min"
    series = pd.Series(
        data=3,
        index=pd.DatetimeIndex(
            pd.date_range(
                start=pd.Timestamp("2022-06-15 12:03"),
                end=pd.Timestamp("2022-06-15 12:20"),
                freq=freq,
            ),
        ),
    )
    missing_series = series.drop(pd.Timestamp("2022-06-15 12:15"))
    expected_series = series.copy()
    expected_series.loc[pd.Timestamp("2022-06-15 12:15")] = 42
    fixed_series = fill_missing_entries(
        missing_series, sampling_period=60, security_factor=2, fill_value=42,
    )
    pd.testing.assert_series_equal(expected_series, fixed_series, check_freq=False)


def test_fill_missing_entry_frame():
    freq = "1min"
    df = pd.DataFrame(
        data=3,
        index=pd.DatetimeIndex(
            pd.date_range(
                start=pd.Timestamp("2022-06-15 12:03"),
                end=pd.Timestamp("2022-06-15 12:20"),
                freq=freq,
            ),
        ),
        columns=["a", "b"],
    )
    missing_frame = df.drop(pd.Timestamp("2022-06-15 12:15"))
    expected_frame = df.copy()
    expected_frame.loc[pd.Timestamp("2022-06-15 12:15")] = 42
    fixed_frame = fill_missing_entries(
        missing_frame, sampling_period=60, security_factor=2, fill_value=42,
    )
    pd.testing.assert_frame_equal(expected_frame, fixed_frame, check_freq=False)


def test_fill_data_holes():
    """Test the fill data holes function"""
    freq = "1min"
    series = pd.Series(
        data=3,
        index=pd.DatetimeIndex(
            pd.date_range(
                start=pd.Timestamp("2022-06-15 12:03"),
                end=pd.Timestamp("2022-06-15 12:20"),
                freq=freq,
            ),
        ),
    )
    missing_series = series.drop(pd.Timestamp("2022-06-15 12:15"))
    expected_series = series.copy()
    expected_series.loc[pd.Timestamp("2022-06-15 12:15")] = 42
    fixed_series = fill_data_holes(
        missing_series, method="median", security_factor=2, fill_value=42,
    )
    pd.testing.assert_series_equal(expected_series, fixed_series, check_freq=False)



"""Analyzes the relation between energy consumption and themperature."""

from .daily_analysis import (
    AutoCategoricalThermoSensitivity,
    DailyCategoricalThermoSensitivity,
    DayOfWeekCategoricalThermoSensitivity,
)
from .thermosensitivity import (
    CategoricalThermoSensitivity,
    ThermoSensitivity,
)



import numpy as np
import pandas as pd

from .. import keywords as EATK
from ..timeseries.create.from_intervals import flatten_and_fill


# =============================================================================
# flatten overconsumption
# =============================================================================
def test_flatten_empty():
    """Empty dataframe is flattened as empty"""
    assert flatten_and_fill(pd.DataFrame()).empty


def test_flatten_one_interval():
    """Check that flattening is OK in one row table."""
    start = pd.Timestamp("2020-02-29")
    end = start + pd.Timedelta("1min")
    test_data = pd.DataFrame(
        [[start, end, 42.0]], columns=[EATK.start_f, EATK.end_f, "test"],
    )
    flat_consumption = flatten_and_fill(
        test_data, start_f=EATK.start_f, end_f=EATK.end_f, time_f=EATK.time_f,
    )
    expect_index = pd.DatetimeIndex([start, end], name=EATK.time_f)
    expect_series = pd.DataFrame(
        np.array([42.0, np.nan]).reshape(2, 1),
        index=expect_index,
        columns=["test"],
    )
    pd.testing.assert_frame_equal(
        expect_series, flat_consumption, check_dtype=False, check_freq=False,
    )


def test_flatten():
    """Check that flattening looks OK."""
    time_f = "timestamp"
    end_f = "end_custom"
    start_f = "start_custom"
    begin = pd.Timestamp("2022-12-05 18:00:00")
    starts = pd.date_range(start=begin, freq="10min", periods=3, name=start_f)
    targets = pd.date_range(start=begin, freq="5min", periods=6, name=time_f)
    data = {
        start_f: starts,
        end_f: starts + pd.Timedelta("5min"),
        "col_str": ["toto"] * 3,
        "col_float": np.arange(0.0, 6.0, 2.0),
    }
    table = pd.DataFrame.from_dict(data)
    # Default filling
    flat_expect = pd.DataFrame.from_dict(
        {
            "col_str": ["toto"] * 6,
            "col_float": np.arange(0, 6),
        },
    )
    flat_expect.index = targets
    flat_expect.iloc[[1, 3, 5], :] = np.nan
    flat_table = flatten_and_fill(table, end_f=end_f, start_f=start_f)
    pd.testing.assert_frame_equal(flat_expect, flat_table, check_freq=False)
    # explicit filling
    flat_filled_table = flatten_and_fill(
        table, fill_values={"col_str": "tata"}, end_f=end_f, start_f=start_f,
    )
    flat_expect["col_str"] = ["toto", "tata"] * 3
    pd.testing.assert_frame_equal(
        flat_expect,
        flat_filled_table,
        check_freq=False,
        check_dtype=False,
    )



"""Testing module for rolling_profile"""

from itertools import product

import numpy as np
import pandas as pd
import pytest

from energy_analysis_toolbox.timeseries.profiles.rolling_profile import (
    RollingProfile,
    RollingQuantileProfile,
)


def arrange_multicolumn(n_days, step, r_rows=10):
    """Generate a dataframe similar to what the pivot phase would produce.

    The values increases linearly from ``0`` to ``n_days*r_rows*step`` along the rows.
    This means that the rolling standard deviation is constant.

    Examples
    --------

    .. code-block::
        python

        >>> arrange_multicolumn(3, 2, 4)
                0	1	2
        0	0	2	4
        1	6	8	10
        2	12	14	16
        3	18	20	22

    """
    return pd.DataFrame(
        np.arange(0, n_days * r_rows * step, step).reshape(r_rows, n_days),
    )


def arrange_days_multicolumn(n_days, step, r_rows=10):
    """Generate a dataframe similar to what the pivot phase would produce.

    The values increases linearly from ``0`` to ``n_days*step`` along the rows.
    This means that the all rolling aggregations are constant.

    Examples
    --------

    .. code-block::
        python

        >>> arrange_days_multicolumn(3, 2, 4)
                0	1	2
        0	0	2	4
        1	0	2	4
        2	0	2	4
        3	0	2	4

    """
    one_row = np.arange(0, n_days * step, step)
    return pd.DataFrame(np.array([one_row] * (r_rows)).reshape(r_rows, n_days))


def expected_constant_df(value, index, window_size):
    """Return a constant dataframe padded with Nan.

    The padding correspond to a centered rolling window with
    a minimum number of rows equal to ``window_size``.

    Parameters
    ----------
    value : Scalar
        the value to set
    index : Index or array-like
        the index of the returned DataFrame
    window_size : int | str | timedelta | ...
        the window size passed to ``pd.rolling``,
        must be consistent with the index type.

    Returns
    -------
    pd.DataFrame
        the expected Dataframe, with one column named ``"value"``

    """
    expected_df = pd.DataFrame(data=value, index=index, columns=["value"])
    return expected_df.rolling(
        window_size,
        center=True,
    ).median()


@pytest.mark.parametrize(
    "n_rows, step, window_size, ddof",
    product(
        [1, 3, 4],  # number of columns
        [1, 2, 3],  # step in the data range
        [3, 4, 7],  # size of the rolling window
        [0, 1],  # DDoF in the STD computation
    ),
)
def test_windowed_rolling_agg_many_combinations(n_rows, step, window_size, ddof):
    """Test the computation of the standard deviation on a rolling window.
    The dataframe generated as several number of columns (1 to 4),
    with data a np.arange with several step values,
    rolling window size and ddof is also tested.

    The expected value is always the same, as the std isn't affected
    by the rising mean value.
    """
    df = arrange_multicolumn(n_rows, step=step)
    expected_std = df.iloc[0:window_size, :].to_numpy().ravel().std(ddof=ddof)
    extected_df = expected_constant_df(expected_std, df.index, window_size)
    rolling_profiler = RollingProfile(
        window=window_size, aggregation=lambda x: np.std(x, ddof=ddof),
    )
    results = rolling_profiler.windowed_rolling_agg(df)
    pd.testing.assert_frame_equal(extected_df, results)


@pytest.mark.parametrize("n_days, step, window_size, freq_minutes", [[3, 2, 5, 6]])
def test_mean_offset(n_days, step, window_size, freq_minutes):
    """Check that the option to add the mean profile to the results is effective when toggled"""
    n_rows = int(24 * 60 / freq_minutes)
    data = arrange_days_multicolumn(n_days, step, n_rows)
    df = pd.DataFrame(
        data={"value": data.to_numpy().T.ravel()},
        index=pd.date_range(
            start="2022-09-23 00:00",
            periods=n_rows * n_days,
            freq=f"{freq_minutes}min",
        ),
    )
    expected_value = np.mean(data.iloc[0, :]) + np.std(data.iloc[0, :])
    expected_time = pd.date_range(
        start="2022-09-23 00:00",
        end="2022-09-23 23:59",
        freq=f"{freq_minutes}min",
    )
    expected_df = expected_constant_df(expected_value, expected_time, window_size)
    rolling_profiler = RollingProfile(
        window=window_size, aggregation=np.std, as_mean_offset=True,
    )
    results = rolling_profiler.compute(df, time=expected_time[0])
    pd.testing.assert_frame_equal(expected_df, results, check_freq=False)


@pytest.mark.parametrize(
    "n_days, step, window_size, q",
    product(
        [1, 3, 4],  # number of columns
        [1, 2, 3],  # step in the data range
        [3, 4, 7],  # size of the rolling window
        [0.5, 0.8],  # DDoF in the STD computation
    ),
)
def test_RollingQuantileProfile_many_combinations(n_days, step, window_size, q):
    """Test that the rolling Quantile works for many combinations.

    The parameter tested are :

    - number of days
    - window_size
    - quantile computed
    - step of the data generation

    """
    freq_minutes = 6
    r_rows = int(n_days * 24 * 60 / freq_minutes)
    data = (
        np.array(np.arange(0, n_days * step, step).tolist() * (r_rows // n_days))
        .reshape(-1, n_days)
        .T.ravel()
    )
    index_name = "timestamp_with_changes"
    index = pd.DatetimeIndex(
        data=pd.date_range(
            start="2022-09-23 00:00",
            periods=r_rows,
            freq=f"{freq_minutes}min",
            name=index_name,
        ),
    )
    df = pd.DataFrame(data, index=index, columns=["value"])
    quantile_profile = RollingQuantileProfile(window_size, threshold_quantile=q)
    results = quantile_profile.compute(df, time=df.index[0])
    expected_value = np.quantile(
        np.array(np.arange(0, n_days * step, step).tolist() * (window_size)),
        q=q,
    )
    expected_time = pd.date_range(
        start="2022-09-23 00:00",
        end="2022-09-23 23:59",
        freq=f"{freq_minutes}min",
        name=index_name,
    )
    expected_df = expected_constant_df(expected_value, expected_time, window_size)
    pd.testing.assert_frame_equal(expected_df, results, check_freq=False)


@pytest.mark.parametrize(
    "n_days, freq, tz",
    product(
        [1, 3, 4],  # number of days
        ["1min", "30min", "3h"],  # timesteps of the data
        ["UTC", None, "Europe/Paris", "Asia/Katmandu"],  # the tz
    ),
)
def test_daily_pivot_many_combinations(n_days, freq, tz):
    """Test that daily_pivot works"""
    start = pd.Timestamp("2023-06-21 00:00:00", tz=tz)
    end = start + pd.DateOffset(days=n_days, milliseconds=-1)
    index = pd.date_range(start=start, end=end, freq=freq)
    history = pd.DataFrame(data=np.ones(len(index)), index=index, columns=["value"])
    result = RollingProfile(None, None).daily_pivot(history)
    assert result.shape[1] == n_days
    pd.testing.assert_index_equal(
        result.index,
        pd.timedelta_range(0, "1day", freq=freq, closed="left", name="time"),
    )


def test_daily_pivot_duplicated_index():
    """Test that daily_pivot works"""
    n_days = 3
    freq = "6min"
    tz = "UTC"
    start = pd.Timestamp("2023-06-21 00:00:00", tz=tz)
    end = start + pd.DateOffset(days=n_days, milliseconds=-1)
    index = pd.date_range(start=start, end=end, freq=freq)
    index = index.append(index[5:7])
    history = pd.DataFrame(data=np.ones(len(index)), index=index, columns=["value"])
    result = RollingProfile(None, None).daily_pivot(history)
    assert result.shape[1] == n_days
    pd.testing.assert_index_equal(
        result.index,
        pd.timedelta_range(0, "1day", freq=freq, closed="left", name="time"),
    )



"""Finds the overconsumption from a power series and a threshold."""

import pandas as pd

from energy_analysis_toolbox.power import basics as power
from energy_analysis_toolbox.timeseries.extract_features import intervals_over


def from_power_threshold(
    power_series: pd.Series,
    overshoot_tshd: pd.Series | float,
    reference_energy_tshd: pd.Series | float | None = None,
) -> pd.DataFrame:
    """Return a table of overconsumption where power_series is above overshoot_tshd.

    The function also computes the overshoot energy as the energy of the difference
    between ``power_series`` and ``reference_energy_tshd`` during the overshoot
    overconsumption.

    Parameters
    ----------
    power_series : pd.Series
        A timeseries of power measures in (W). The power in each element is
        the averaged power during ``[ti, ti+1[`` where ``ti`` and ``ti+1`` are the
        indices of the considered and next elements.
    overshoot_tshd : pd.Series or float
        The threshold in (W) over which the power is considered as
        over-consumption. In case a series is given, it should have the
        same index as ``power_series``.
    reference_energy_tshd : pd.Series or float or None
        A power in (W) to be subtracted from the power series in order
        to compute an "overshoot energy" for each interval. The
        default is |None| in which case ``overshoot_tshd`` is used.
        In case a series is given, it should have the same index as
        ``power_series``.

    Returns
    -------
    pd.DataFrame :
        A table of overconsumption  with the following columns :

        - ``start`` : timestamp of the first instant of an overshoot interval;
        - ``end`` : timestamp of the first instant after an overshoot interval;
        - ``duration`` : duration in (s) of the overshoot interval. This is the
          difference in (s) between the start and end bounds.
        - ``energy`` : the energy associated to the difference between ``power_series``
          and ``reference_energy_tshd`` during the ``[start, end[`` interval.


    .. note:: Why use two thresholds ?

        When looking for overconsumption, the threshold defining the "anomaly" of
        the power may not be the reference VS which the overconsumption is computed.
        E.g., the overconsumption may be computed VS the average consumption while the
        abnormal overconsumption may be identified using variability-related thresholds.

        This function provides both location and "overshoot-energy" computation for
        convenience.

    """
    if reference_energy_tshd is None:
        reference_energy_tshd = overshoot_tshd
    # Find overshoots of power threshold
    if isinstance(overshoot_tshd, pd.Series):
        power_series_aligned, overshoot_tshd_aligned = power_series.align(
            overshoot_tshd,
        )
    else:
        power_series_aligned, overshoot_tshd_aligned = (
            power_series,
            pd.Series(overshoot_tshd, index=power_series.index),
        )
    intervals_overshoot = intervals_over(
        power_series_aligned > overshoot_tshd_aligned, 0.5,
    )
    if intervals_overshoot.empty:
        intervals_overshoot["duration"] = []
        intervals_overshoot["energy"] = []
    else:
        intervals_overshoot["duration"] = (
            intervals_overshoot["end"] - intervals_overshoot["start"]
        ).dt.total_seconds()
        # Compute energy criterion associated to overshoots
        if isinstance(reference_energy_tshd, pd.Series):
            power_series_aligned, reference_energy_tshd_aligned = power_series.align(
                reference_energy_tshd,
            )
        else:
            power_series_aligned, reference_energy_tshd_aligned = (
                power_series,
                pd.Series(reference_energy_tshd, index=power_series.index),
            )
        intervals_overshoot["energy"] = power.integrate_over(
            intervals_overshoot, power_series - reference_energy_tshd_aligned,
        )
    return intervals_overshoot



"""Defines custom errors for time-series."""
from .base import EATError


class EATInvalidTimeseriesError(EATError):
    """A timeseries is inconsistent.

    This exception is used to notify a problem regarding the consistency of
    a timeseries definition.

    Derived classes may be inherited in order to deal with more specific problems.
    """


class EATUndefinedTimestepError(EATInvalidTimeseriesError):
    """A timestep is undefined in an interval-sampled timeseries.

    A timestep duration is undefined, either explicitly or implicitly in a
    timeseries which contain values defined "over overconsumption" (VS at
    instantaneous timestamps).
    """


class EATInvalidTimestepDurationError(EATInvalidTimeseriesError):
    """An invalid timestep duration has been passed for a timeseries.

    This exception is used when a timestep duration is encountered which does
    not comply with implicit or explicit assumptions about the possible timestep
    durations in a timeseries. Such assumptions could be :

    - durations are >=0,
    - durations have a specific value,
    - etc.

    """



"""Compute time-derivatives of physical values in a timeseries.

Available derivative calculations are:

- :py:func:`time_derivative_fwd`
- :py:func:`time_derivative_second`

These implementations use basic forward time-derivatives schemes, more suited
for the interest of signal analysis in the draw-detection than for numerical
simulation.

"""

import numpy as np
import pandas as pd


def time_derivative_fwd(
    timeseries: pd.Series,
) -> pd.Series:
    r"""Return the forward 1st-order time-derivative of a time-series.

    Parameters
    ----------
    timeseries : pd.Series
        A timeseries.

    Returns
    -------
    pd.Series :
        The timeseries of forward time-derivative of the input series


    .. important::

        The function assumes that the provided timeseries has at least 3
        elements and that all the elements are ordered following chronological
        order.


    Notes
    -----
    Let `f` be a function of time sampled for `N` values indexed by
    chronological order. Assume `N > 2`.

    The derivative values of the timeseries representing the ordered sequence
    of :math:`(f(t_{i}))_{i \\in [ 1.. N ]` are computed using the following
    formulas :

    .. math::

        \\frac{df}{dt}(t_i) = \\frac{f(t_{i+1}) - f(t_{i})}{t_{i+1} - t_{i}}
        \\forall i \\in [1.. N -1]

        \\frac{df}{dt}(t_N) = \\frac{f(t_{N}) - f(t_{N-1})}{t_{N} - t_{N - 1}}


    .. warning::

        This first order forward scheme is very crude. Its used is dicouraged in
        physical numerical simulation, in which the accuracy and stability of the
        scheme are of great importance.


    Example
    -------
    .. code-block:: python

        >>> import pandas as pd
        >>> import numpy as np
        >>> time = pd.date_range(pd.Timestamp("2021-03-14"), freq='1S', periods=10)
        >>> values = np.arange(0, 10)**2
        >>> series = pd.Series(values, index=time)
        >>> series
        2021-03-14 00:00:00     0
        2021-03-14 00:00:01     1
        2021-03-14 00:00:02     4
        ...
        2021-03-14 00:00:07    49
        2021-03-14 00:00:08    64
        2021-03-14 00:00:09    81
        Freq: S, dtype: int64
        >>> time_derivative_fwd(series)
        2021-03-14 00:00:00     1.0
        2021-03-14 00:00:01     3.0
        2021-03-14 00:00:02     5.0
        ...
        2021-03-14 00:00:07    15.0
        2021-03-14 00:00:08    17.0
        2021-03-14 00:00:09    17.0
        Freq: S, dtype: float64


    """
    dts = (timeseries.index - timeseries.index[0]).total_seconds()
    grads = np.empty(timeseries.index.size)
    grads[0:-1] = np.ediff1d(timeseries.values) / np.ediff1d(dts)
    grads[-1] = (timeseries.to_numpy()[-1] - timeseries.to_numpy()[-2]) / (
        dts[-1] - dts[-2]
    )
    return pd.Series(grads, index=timeseries.index)


def time_derivative_second(
    timeseries: pd.Series,
) -> pd.Series:
    """Return the forward second order time-derivative of a time-series.

    Parameters
    ----------
    timeseries : pd.Series
        A timeseries.

    Returns
    -------
    pd.Series :
        The timeseries of double time-derivative of the input series.


    .. important::

        The function assumes that the provided timeseries has at least 3
        elements and that all the elements are ordered following chronological
        order.


    Notes
    -----
    The second order time-derivative of the input timeseries is obtained by
    applying twice the formula used in :py:func:`time_derivative_fwd`.

    .. warning::

        This time-derivative scheme is very crude. Its used is dicouraged in
        physical numerical simulation, in which the accuracy and stability of the
        scheme are of great importance.

        In particular, due to the way the derivative is computed, the last two
        values of the series are always 0, as seen in the example below.


    Example
    -------

    .. code-block:: python

        >>> import pandas as pd
        >>> import numpy as np
        >>> time = pd.date_range(pd.Timestamp("2021-03-14"), freq='1S', periods=10)
        >>> values = np.arange(0, 10)**2
        >>> series = pd.Series(values, index=time)
        >>> series
        2021-03-14 00:00:00     0
        2021-03-14 00:00:01     1
        2021-03-14 00:00:02     4
        ...
        2021-03-14 00:00:07    49
        2021-03-14 00:00:08    64
        2021-03-14 00:00:09    81
        Freq: S, dtype: int64
        >>> time_derivative_second(series)
        2021-03-14 00:00:00    2.0
        2021-03-14 00:00:01    2.0
        2021-03-14 00:00:02    2.0
        ...
        2021-03-14 00:00:07    2.0
        2021-03-14 00:00:08    0.0
        2021-03-14 00:00:09    0.0
        Freq: S, dtype: float64

    """
    dts = (timeseries.index - timeseries.index[0]).total_seconds()
    grads = np.empty(timeseries.index.size)
    grads[0:-1] = np.ediff1d(timeseries.values) / np.ediff1d(dts)
    grads[-1] = (timeseries.to_numpy()[-1] - timeseries.to_numpy()[-2]) / (
        dts[-1] - dts[-2]
    )
    grads2 = np.empty(timeseries.index.size)
    grads2[0:-1] = np.ediff1d(grads) / np.ediff1d(dts)
    grads2[-1] = (grads[-1] - grads[-2]) / (dts[-1] - dts[-2])
    return pd.Series(grads2, index=timeseries.index)



"""Add an accessor to pandas.DataFrame and pandas.Series for the computation toolbox.

**In order for the accessors to be accessible, this module must be imported.**
The module is not imported by default when importing |eat|::

    import energy_analysis_toolbox as et
    import energy_analysis_toolbox.pandas


Series
------
A pandas series convey only one steam of data. It can be a timeseries of power, energy,
temperature, etc.

Hence, the functionalities of the computation toolbox are limited to the following
assumptions:

- The index is a datetime index
- The values are numeric
- There is no missing values, i.e. the index is complete, and the value correspond to
  the interval between the index and the next index.
- If not provided as an argument, the last timestep duration is assumed to be
  the same as the previous one.

Knowing this, if the accessor has been enabled, it becomes available on any
pandas.Series with name `eat`. Operations such as the following becom possible::

    my_series.eat.to_freq('1h', method='volume_conservative')
    a_power_series.eat.to_energy()


More examples in :doc:`/user_guide/using_the_accessor`.

Examples of use
~~~~~~~~~~~~~~~

If the accessor has been enabled, it becomes available on any pandas.DataFrame with
name `eat`. Operations such as the following becom possible::

    power_data.eat.power_to_freq('1h')
    power_data.eat.power_to_energy()


More examples in :doc:`/user_guide/using_the_accessor`.

"""

from typing import Literal

import pandas as pd

from energy_analysis_toolbox.timeseries.resample._facade import resampling_methods

from . import (
    energy,
    power,
    timeseries,
)


@pd.api.extensions.register_series_accessor("eat")
class EATAccessorSeries:
    """Define a new namespace for the computation toolbox on pandas.Series."""

    def __init__(
        self,
        data: pd.Series,
    ) -> None:
        if not isinstance(data, pd.Series):
            err = f"Expected the input to be a Series, but got {type(data)}."
            raise TypeError(err)
        self._obj = data

    def to_energy(
        self,
        *args,
        **kwargs,
    ) -> pd.Series:
        """Convert a power series to an energy series.

        See :func:`energy_analysis_toolbox.power.to_energy` for details.

        Returns
        -------
        pd.Series
            An energy series.

        """
        return power.to_energy(
            self._obj,
            *args,
            **kwargs,
        )

    def to_power(
        self,
        *args,  # noqa:ARG002
        **kwargs,  # noqa:ARG002
    ) -> pd.Series:
        """Convert an energy series to a power series.

        See :func:`energy_analysis_toolbox.energy.to_power` for details.

        Returns
        -------
        pd.Series
            A power series.

        """
        err = "to_power method is not implemented yet."
        raise NotImplementedError(err)

    def power_to_freq(
        self,
        *args,
        **kwargs,
    ) -> pd.Series:
        """Resample a power series to a fixed frequency.

        See :func:`energy_analysis_toolbox.power.to_freq` for details.

        Returns
        -------
        pd.Series
            A power series resampled to a fixed frequency.

        """
        return power.to_freq(
            self._obj,
            *args,
            **kwargs,
        )

    def energy_to_freq(
        self,
        *args,
        **kwargs,
    ) -> pd.Series:
        """Resample an energy series to a fixed frequency.

        See :func:`energy_analysis_toolbox.energy.to_freq` for details.

        Returns
        -------
        pd.Series
            An Energy series resampled to a fixed frequency.

        """
        return energy.to_freq(self._obj, *args, **kwargs)

    def to_freq(
        self,
        freq: str | pd.Timedelta | None,
        origin: None | Literal["floor", "ceil"] | pd.Timestamp = None,
        last_step_duration: float | None = None,
        method: resampling_methods = "piecewise_affine",
        **kwargs,
    ) -> pd.Series:
        """Resample a series to a fixed frequency with various strategies.

        See :func:`energy_analysis_toolbox.timeseries.resample.to_freq` for details.

        Parameters
        ----------
        freq : str, pd.Timedelta
            the freq to which the series is resampled. Must be a valid
            pandas frequency.
        origin : {None, 'floor', 'ceil', pd.Timestamp}
            What origin should be used for the target resampling range. The following
            values are possible :

            - |None| : the default. Use the first index as the data a starting point.
            - ``'floor'`` : use the first index of the data, floored to the passed
              ``freq`` resolution.
            - ``'ceil'`` : use the first index of the data, ceiled to the passed
              ``freq`` resolution.
            - a ``pd.Timestamp`` : use the passed timestamp as starting point. The
              code tries to localize the value to the timezone of the first index in
              the data. Accordingly :

            * if the passed value is time-naive, it is localized to the timezone
              of the data;
            * if the data is time-naive, the timezone of the passed value is ignored
              and it is processed as if it were time-naive.

        last_step_duration : float, optional
            the duration of the last step of the resampling in (s).
            If |None|, the duration of the former-last time-step is used.
            Used to deduce the end of the resampling range.

        method : str or callable, optional
            Method used to interpolate the values of the resampled series. The accepted
            values are:

            * 'piecewise_affine': uses :py:func:`.piecewise_affine`, assume the values
              a straight line between two points. The default method.
            * 'piecewise_constant': uses :py:func:`.piecewise_constant`, assume the
              values constante until the next point.
            * 'volume_conservative': uses :py:func:`.volume_to_freq`, conserve
              quantity of the values. Best to use it for energy timeseries.
            * 'flow_rate_conservative': uses :py:func:`.flow_rate_to_freq`, conserve
              the values time the duration between two points. Best to use it for
              power timeseries.

            If a callable is passed, it must take a :py:class:`pandas.Series` as first
            argument and a :py:class:`pandas.DatetimeIndex` as second argument.
            See the interface of :py:func:`piecewise_affine` function.
            The default is 'piecewise_affine'.

        Returns
        -------
        pd.Series
            A series resampled to a fixed frequency.

        """
        return timeseries.resample.to_freq(
            timeseries=self._obj,
            freq=freq,
            origin=origin,
            last_step_duration=last_step_duration,
            method=method,
            **kwargs,
        )

    def intervals_over(
        self,
        *args,
        **kwargs,
    ) -> pd.DataFrame:
        """Detect intervals over a threshold.

        See :func:`energy_analysis_toolbox.timeseries.extract_features.intervals_over`
        for details.

        Returns
        -------
        pd.DataFrame
            The intervals over the threshold.

        """
        return timeseries.extract_features.intervals_over(
            self._obj,
            *args,
            **kwargs,
        )

    def timestep_durations(
        self,
        *args,
        **kwargs,
    ) -> pd.Series:
        """Return the series of timestep durations of a timeseries.

        See :func:`energy_analysis_toolbox.timeseries.timestep_durations` for details.

        Returns
        -------
        pd.Series
            The duration of each timestep.

        """
        return timeseries.extract_features.timestep_durations(
            self._obj,
            *args,
            **kwargs,
        )

    def fill_data_holes(
        self,
        *args,
        **kwargs,
    ) -> pd.Series:
        """Fill the holes in a timeseries.

        See :func:`energy_analysis_toolbox.timeseries.fill_data_holes` for details.

        Returns
        -------
        pd.Series
            The timeseries with the holes filled.

        """
        return timeseries.resample.fill_data_holes(
            self._obj,
            *args,
            **kwargs,
        )


@pd.api.extensions.register_dataframe_accessor("eat")
class EATAccessorDataFrame:
    """Define a new namespace for the computation toolbox on pandas.Series."""

    def __init__(
        self,
        data: pd.DataFrame,  # noqa:ARG002
    ) -> None:
        err = """EAT accessor for DataFrame is not implemented yet."""
        raise NotImplementedError(err)



"""Transforms indices of a time series to a new index according to a given function."""

from contextlib import suppress
from typing import TypeVar

import numpy as np
import pandas as pd
from pytz import BaseTzInfo
from scipy.stats import gaussian_kde, mode

from energy_analysis_toolbox.errors import EATUndefinedTimestepError
from energy_analysis_toolbox.timeseries.extract_features.basics import (
    index_to_timesteps,
    timestep_durations,
)

T = TypeVar("T", pd.Series, pd.DataFrame)


def tz_convert_or_localize(
    timeseries: pd.Series | pd.DataFrame,
    tz: str | BaseTzInfo | None,
) -> pd.Series | pd.DataFrame:
    """Assign the requested timezone to the index of a timeseries.

    Parameters
    ----------
    timeseries : pd.Series or pd.DataFrame
        Timeseries to convert.
    tz : str or pytz.timezone or None
        Timezone to assign to the index of the timeseries.

    Returns
    -------
    pd.Series
        Timeseries with the requested timezone assigned to its index.

    .. note::

        This function is just syntactic sugar to avoid dealing with the `TypeError`
        when applying `tz_convert` to a time-naive timeseries.


    .. important::

        When localizing a time-naive timeseries, the `ambiguous` and `nonexistent`
        arguments are set to `True` and `'NaT'` respectively. This means that
        ambiguous times are localized to the beginning of the DST period
        and non-existent times are converted to `'NaT'`.

    """
    try:
        return timeseries.tz_convert(tz)
    except TypeError:
        return timeseries.tz_localize(tz, ambiguous=True, nonexistent="NaT")


def index_to_freq(
    index: pd.DatetimeIndex,
    freq: str | pd.Timedelta | None,
    origin: str | pd.Timestamp | None = None,
    last_step_duration: float | None = None,
) -> pd.DatetimeIndex:
    """Return the expected index from resampling a time series to a given frequency.

    Parameters
    ----------
    index : pd.DatetimeIndex
        the index of the data to resample
    freq : str, pd.Timedelta
        the freq to which the series is resampled. Must be a valid
        pandas frequency.
    origin : {None, 'floor', 'ceil', pd.Timestamp}
        What origin should be used for the target resampling range. The following
        values are possible :

        - |None| : the default. Use the first index as the data a starting point.
        - ``'floor'`` : use the first index of the data, floored to the passed
          ``freq`` resolution.
        - ``'ceil'`` : use the first index of the data, ceiled to the passed
          ``freq`` resolution.
        - a ``pd.Timestamp`` : use the passed timestamp as starting point. The
          code tries to localize the value to the timezone of the first index in
          the data. Accordingly :

          * if the passed value is time-naive, it is localized to the timezone
            of the data;
          * if the data is time-naive, the timezone of the passed value is ignored
            and it is processed as if it were time-naive.

    last_step_duration : float, optional
        the duration of the last step of the resampling in (s).
        If |None|, the duration of the former-last time-step is used.
        Used to deduce the end of the resampling range.

    Returns
    -------
    pd.DatetimeIndex
        The resulting index of the resampling. Empty if the passed index is empty.

    """
    if index.empty:
        return pd.DatetimeIndex([], name=index.name, tz=index.tz, freq=freq)
    if origin is None:
        start = index[0]
    elif origin == "floor":
        start = index[0].floor(freq)
    elif origin == "ceil":
        start = index[0].ceil(freq)
    else:
        start = pd.Timestamp(origin)
        try:
            start = start.tz_localize(index.tz)
        except TypeError:
            try:
                start = start.tz_convert(index.tz)
            except TypeError:
                warn = (
                    "The passed origin could not be localized or converted to the "
                    "timezone of the original index. It is processed as if it were "
                    "time-naive."
                )
                raise Warning(warn) from None
    if last_step_duration is None:
        try:
            last_step_duration = (index[-1] - index[-2]).seconds
        except IndexError:
            err = (
                "The last step duration could not be determined from the index."
                " Please provide it explicitly."
            )
            raise EATUndefinedTimestepError(err) from None
    actual_end = index[-1] + pd.Timedelta(seconds=last_step_duration)
    return pd.date_range(
        start=start,
        end=actual_end,
        freq=freq,
        inclusive="left",
        name=index.name,
    )


def estimate_timestep(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
    method: str = "median",
) -> float:
    """Return an estimation of the sampling period of a time series.

    .. note::
        Each method has its own advantages and drawbacks. The best method
        depends on the data. For instance:

        - if the data is regularly spaced, the ``mode`` is the best choice.
        - if the data is irregularly spaced, the ``kde`` is the best choice.
        - the ``median`` is not sensitive to outliers, and is a good choice if
            the data is irregularly spaced and has outliers.
        - the ``mean`` is almost never a good choice.

    Parameters
    ----------
    data : pd.Series, pd.DataFrame, pd.DatetimeIndex
        the data to analyse. Must have (or be) a DatetimeIndex.
    method : {'mean', 'median', 'mode', 'kde'}, optional
        the method used to compute the expected timestep. Defaults to 'median'.

    Returns
    -------
    float
        the expected timestep of the data in (s).

    Raises
    ------
    ValueError
        If the method is not one of {'mean', 'median', 'mode', 'kde'}.


    .. seealso::
        - :func:`median_time_step`
        - :func:`mean_time_step`
        - :func:`mode_time_step`
        - :func:`max_kde_time_step`

    """
    if method == "mean":
        return mean_time_step(data)
    if method == "median":
        return median_time_step(data)
    if method == "mode":
        return mode_time_step(data)
    if method == "kde":
        return max_kde_time_step(data)
    err = "method must be one of {'mean', 'median', 'mode', 'kde'}"
    raise ValueError(err)


def median_time_step(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
) -> float:
    """Return the median timestep of a time series.

    Parameters
    ----------
    data : pd.Series, pd.DataFrame, pd.DatetimeIndex
        the data to analyse. Must have (or be) a DatetimeIndex.

    Returns
    -------
    float
        the median timestep of the data in (s).


    .. seealso::
        - :func:`estimate_timestep`
        - :func:`mode_time_step`

    """
    data = data_to_datetimeindex(data)
    timesteps = index_to_timesteps(data)
    return np.median(timesteps)


def mean_time_step(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
) -> float:
    """Return the mean timestep of a time series.

    Parameters
    ----------
    data : pd.Series, pd.DataFrame, pd.DatetimeIndex
        the data to analyse. Must have (or be) a DatetimeIndex.

    Returns
    -------
    float
        the mean timestep of the data in (s).


    .. seealso::
        - :func:`estimate_timestep`
        - :func:`mode_time_step`

    """
    data = data_to_datetimeindex(data)
    timesteps = index_to_timesteps(data)
    return np.mean(timesteps)


def mode_time_step(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
) -> float:
    """Return the mode timestep of a time series.

    .. warning:: The mode is the most frequent value. If there are several
        values with the same frequency, the first one is returned.
        If the values vary slightly around a central value, the mode
        is not representative of the data.

    Parameters
    ----------
    data : pd.Series, pd.DataFrame, pd.DatetimeIndex
        the data to analyse. Must have (or be) a DatetimeIndex.

    Returns
    -------
    float
        the mode timestep of the data in (s).


    .. seealso::
        - :func:`estimate_timestep`
        - :func:`max_kde_time_step`

    """
    data = data_to_datetimeindex(data)
    timesteps = index_to_timesteps(data)
    return mode(timesteps, nan_policy="omit").mode


def max_kde_time_step(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
) -> float:
    """Return the maximum probable timestep of a time series.

    .. note:: It differs from the Mode as the distribution is first
       estimated using a KDE. Then, the max of this distribution is
       used.

    .. warning:: The KDE cannot be estimated if the data is regularly
        spaced. In this case, use another method.

    Parameters
    ----------
    data : pd.Series, pd.DataFrame, pd.DatetimeIndex
        the data to analyse. Must have (or be) a DatetimeIndex.

    Returns
    -------
    float
        the mode timestep of the data in (s).


    .. seealso::
        - :func:`estimate_timestep`
        - :func:`median_time_step`
        - :func:`mode_time_step`

    """
    data = data_to_datetimeindex(data)
    timesteps = index_to_timesteps(data)
    kde = gaussian_kde(timesteps)
    no_samples = 50
    samples = np.linspace(min(timesteps), max(timesteps), no_samples)
    probs = kde.evaluate(samples)
    maxima_index = probs.argmax()
    return samples[maxima_index]


def data_to_datetimeindex(
    data: pd.Series | pd.DataFrame | pd.DatetimeIndex,
) -> pd.DatetimeIndex:
    """Convert the data to DatetimeIndex.

    Used to allow the use of the same functions for Series, DataFrame and
    DatetimeIndex.

    Parameters
    ----------
    data : pd.Series | pd.DataFrame | pd.DatetimeIndex
        A pandas object

    Returns
    -------
    pd.DatetimeIndex
        Return data if already an index, else the index of the data

    Raises
    ------
    ValueError
        If the data cannot be converted to pandas.DateTimeIndex

    """
    with suppress(AttributeError):
        data = data.index
    if not isinstance(data, pd.DatetimeIndex):
        err = "The data cannot be converted to pandas.DateTimeIndex"
        raise TypeError(err)
    return data


def fill_missing_entries(
    data: pd.Series | pd.DataFrame,
    sampling_period: float,
    security_factor: float = 2,
    fill_value: float = pd.NA,
) -> pd.Series | pd.DataFrame:
    """Fill the data with new entries where the interval is too long.

    .. note:: The duration between the last new entry of a hole and the next
        (existing) entry is less or equal than the sampling_period.

    Parameters
    ----------
    data : pd.Series | pd.DataFrame
        The Data to process. Must have a DatetimeIndex
    sampling_period : float,
        The expected sampling period in (s)
    security_factor : float, optional
        The factor used to determine when a timestep is too long compared to
        the ``sampling_period``, which means that ``sampling_period * security_factor``
        is the maximum duration (excluded) between two entries.
        By default 2.
    fill_value : float, optional
        The value of the newly created entries, by default pd.NA

    Returns
    -------
    pd.Series | pd.DataFrame
        A copy of `data` with new created entries, sorted by index.


    .. seealso::

        - :func:`fill_data_holes`
        - :func:`estimate_timestep`

    """
    durations = timestep_durations(data)
    intervals_to_fill = durations[durations >= sampling_period * security_factor]
    if intervals_to_fill.empty:
        return data
    new_indexes = []
    for index, duration in intervals_to_fill.items():
        number_missing_entries = int(duration // sampling_period)
        tmp_indexes = [
            index + k * pd.Timedelta(seconds=sampling_period)
            for k in range(1, number_missing_entries)
        ]
        new_indexes += tmp_indexes
    missing_index = pd.DatetimeIndex(new_indexes)
    return data.reindex(
        data.index.append(missing_index).sort_values(),
        fill_value=fill_value,
    )


def fill_data_holes(
    data: T,
    method: str = "mode",
    security_factor: float = 2,
    fill_value: float = pd.NA,
) -> T:
    """Return the data with new entries where the interval is too long.

    .. note:: the new indexes are created using the expected timestep determined
        by ``method``. The duration between the last new entry of a hole and the next
        (existing) entry is less or equal than the expected timestep.

    Parameters
    ----------
    data : pd.Series | pd.DataFrame
        The Data to process. Must have a DatetimeIndex
    method : {'mean', 'median', 'mode', 'kde'}, optional
        The method to estimate the expected Frequency, by default "mode".
        See :func:`estimate_timestep` for more details.
    security_factor : float, optional
        The factor used to determine a timestep is too long compared to
        the expected frequency, by default 2.
    fill_value : float, optional
        The value of the newly created entries, by default pd.NA

    Returns
    -------
    pd.Series | pd.DataFrame
        A copy of `data` with new created entries, sorted.


    .. seealso::
        - :func:`fill_missing_entries`
        - :func:`estimate_timestep`

    """
    sampling_period = estimate_timestep(data, method=method)
    return fill_missing_entries(
        data,
        sampling_period,
        security_factor=security_factor,
        fill_value=fill_value,
    )



import pandas as pd
import pytest

from ..errors import EATUndefinedTimestepError
from ..power import to_energy, to_freq


def test_to_freq_no_data():
    """Test case for the to_freq function when the power_series is empty.
    It verifies that the new_power_series is also empty.
    """
    power_series = pd.Series()
    new_freq = "30s"
    new_power_series = to_freq(power_series, new_freq)
    assert new_power_series.empty


def test_to_freq_one_to_many():
    """Test case when the input series has length 1.

    This test case verifies that the `to_freq` function correctly resamples a power series with only one value to a
    higher frequency.
    """
    power_series = pd.Series([1.0], index=[pd.Timestamp("2020-01-01")])
    freq = "30min"
    new_freq = "1min"
    with pytest.raises(EATUndefinedTimestepError):
        new_power_series = to_freq(power_series, new_freq)
    new_power_series = to_freq(
        power_series, new_freq, last_step_duration=pd.Timedelta(freq).seconds,
    )
    assert new_power_series.index.freq == pd.Timedelta(new_freq)
    assert new_power_series.index[0] == pd.Timestamp("2020-01-01")
    assert len(new_power_series) == 30
    assert to_energy(new_power_series).sum() == pd.Timedelta(freq).seconds


def test_to_freq_two_to_many():
    """Test case when the input series has length 2.

    Also test that the last step duration is correctly used.
    """
    power_series = pd.Series(
        [1.0, 1.0],
        index=[pd.Timestamp("2020-01-01"), pd.Timestamp("2020-01-01 00:30:00")],
    )
    freq = "30min"
    new_freq = "1min"
    new_power_series = to_freq(power_series, new_freq)
    assert new_power_series.index.freq == pd.Timedelta(new_freq)
    assert new_power_series.index[0] == pd.Timestamp("2020-01-01")
    assert len(new_power_series) == 60
    assert to_energy(new_power_series).sum() == 2 * pd.Timedelta(freq).seconds
    # last step duration is force to half the new period
    new_power_series = to_freq(
        power_series,
        new_freq,
        last_step_duration=pd.Timedelta(freq).seconds / 2,
    )
    assert len(new_power_series) == 45
    assert to_energy(new_power_series).sum() == 1.5 * pd.Timedelta(freq).seconds


def test_to_freq_many_to_one():
    """Test case when the input series has a duration of 30 minutes but is resample to one element."""
    freq = "1min"
    power_series = pd.Series(
        [1.0] * 30,
        index=pd.date_range(start="2020-01-01", periods=30, freq=freq),
    )
    new_freq = "1h"
    new_power_series = to_freq(power_series, new_freq)
    assert new_power_series.index.freq == pd.Timedelta(new_freq)
    assert new_power_series.index[0] == pd.Timestamp("2020-01-01")
    assert len(new_power_series) == 1
    assert (
        new_power_series.sum() * pd.Timedelta(new_freq).seconds
        == pd.Timedelta(freq).seconds * 30
    )
    # last step duration is force to half the new period
    new_power_series = to_freq(
        power_series,
        new_freq,
        last_step_duration=pd.Timedelta(freq).seconds / 2,
    )
    assert len(new_power_series) == 1
    assert (
        new_power_series.sum() * pd.Timedelta(new_freq).seconds
        == pd.Timedelta(freq).seconds * 29.5
    )


def test_to_freq_one_to_one():
    """Test case when the input series has length 1 and is resampled to one element."""
    freq = "1min"
    power_series = pd.Series(
        [1.0], index=pd.date_range(start="2020-01-01", periods=1, freq=freq),
    )
    new_freq = "1h"
    new_power_series = to_freq(
        power_series, new_freq, last_step_duration=pd.Timedelta(freq).seconds,
    )
    assert new_power_series.index.freq == pd.Timedelta(new_freq)
    assert new_power_series.index[0] == pd.Timestamp("2020-01-01")
    assert len(new_power_series) == 1
    assert (
        new_power_series.sum() * pd.Timedelta(new_freq).seconds
        == pd.Timedelta(freq).seconds
    )



import pandas as pd
import pytest

from ..power.overconsumption.select import (
    by_combined_proportions,
    by_cumulated_proportion,
    by_individual_proportion,
)


def example_intervals():
    """Return 4 overconsumption of overconsumption."""
    s1 = pd.Timestamp("2023-06-21 11:07")
    e1 = pd.Timestamp("2023-06-21 11:14")
    # ------------------------------------
    s2 = pd.Timestamp("2023-06-21 12:00")
    e2 = pd.Timestamp("2023-06-21 13:12")
    # ------------------------------------
    s3 = pd.Timestamp("2023-06-21 13:15")
    e3 = pd.Timestamp("2023-06-21 13:17:30")
    # ------------------------------------
    s4 = pd.Timestamp("2023-06-21 18:42")
    e4 = pd.Timestamp("2023-06-21 18:47")
    # ------------------------------------
    intervals = pd.DataFrame.from_dict(
        {
            "start": [s1, s2, s3, s4],
            "end": [e1, e2, e3, e4],
            "duration": [420.0, 4320.0, 150.0, 300.0],
            "energy": [1500.0, 3000.0, 500.0, 5000.0],  # total 10.000
        },
    )
    return intervals


def test_by_individual_proportions():
    """Check selection by individual proportion on various cases :

    1. proportion equal to the smallest one : all overconsumption selected
    2. intermediate proportion : select only a subset (one here)
    3. proportion higher than the biggest one in overconsumption : empty selection
    4. 3 with custom reference equal to half the total energy : two overconsumption
    """
    intervals = example_intervals()
    # all of them
    case_1 = by_individual_proportion(
        intervals_overshoot=intervals, proportion_tshd=0.05,
    )
    case_1_expect = intervals.copy().sort_values(by="energy", ascending=False)
    case_1_expect["proportion"] = [0.5, 0.3, 0.15, 0.05]
    pd.testing.assert_frame_equal(case_1, case_1_expect)
    # only some
    case_2 = by_individual_proportion(
        intervals_overshoot=intervals, proportion_tshd=0.4,
    )
    case_2_expect = case_1.copy().iloc[0:1]
    pd.testing.assert_frame_equal(case_2, case_2_expect)
    # None
    case_3 = by_individual_proportion(
        intervals_overshoot=intervals, proportion_tshd=0.6,
    )
    # custom reference
    case_4 = by_individual_proportion(
        intervals_overshoot=intervals,
        proportion_tshd=0.6,
        energy_reference=5000.0,
    )
    case_4_expect = case_1_expect.copy().iloc[:2]
    case_4_expect["proportion"] = [1.0, 0.6]
    pd.testing.assert_frame_equal(case_4, case_4_expect)
    assert case_3.empty
    # check not inplace
    with pytest.raises(KeyError):
        intervals["proportion"]


def test_by_cumulated_proportions():
    """Check selection by cumulated proportion on various cases :

    1. proportion equal to 1 : all overconsumption selected
    2. intermediate proportion > biggest: select only a subset (two here)
    3. intermediate proportion < biggest: select biggest overshoot
    4. proportion = 0 : select biggest overshoot anyway
    5. proportion = 0.5 and custom reference equal to twice the total energy :
       select all.

    """
    intervals = example_intervals()
    # all of them
    case_1 = by_cumulated_proportion(intervals_overshoot=intervals, proportion_tshd=1.0)
    case_1_expect = intervals.copy().sort_values(by="energy", ascending=False)
    case_1_expect["cum_energy_prop"] = [0.5, 0.8, 0.95, 1.0]
    pd.testing.assert_frame_equal(case_1, case_1_expect)
    # only some
    case_2 = by_cumulated_proportion(intervals_overshoot=intervals, proportion_tshd=0.6)
    case_2_expect = case_1.copy().iloc[0:2]
    pd.testing.assert_frame_equal(case_2, case_2_expect)
    # only one
    case_3 = by_cumulated_proportion(intervals_overshoot=intervals, proportion_tshd=0.4)
    case_3_expect = case_1.copy().iloc[0:1]
    pd.testing.assert_frame_equal(case_3, case_3_expect)
    # 0 proportion : always at least one value (same as case 3)
    case_4 = by_cumulated_proportion(intervals_overshoot=intervals, proportion_tshd=0)
    pd.testing.assert_frame_equal(case_4, case_3_expect)
    # custom reference
    case_5 = by_cumulated_proportion(
        intervals_overshoot=intervals,
        proportion_tshd=0.5,
        energy_reference=20000.0,
    )
    case_5_expect = case_1_expect.copy()
    case_5_expect["cum_energy_prop"] /= 2
    pd.testing.assert_frame_equal(case_5, case_5_expect)
    # check not inplace
    with pytest.raises(KeyError):
        intervals["cum_energy_prop"]


def test_by_combines_proportions():
    """Check selection by cumulated proportion on various cases.

    Let (pc, pi) be the combined and individual proportions

    1. (1, 0) : all overconsumption selected
    2. (1, 0.1) : drop the smallest one (0.05)
    3. (0.5, 0.1) : keep only the biggest (0.5)
    4. (0, 1) : no interval
    5. (0, 1) and reference is half of total energy : keep only the biggest (0.5)
    """
    intervals = example_intervals()
    # all of them
    case_1 = by_combined_proportions(
        intervals_overshoot=intervals,
        proportion_tshd=1.0,
        proportion_indiv_tshd=0.0,
    )
    case_1_expect = intervals.copy().sort_values(by="energy", ascending=False)
    case_1_expect["cum_energy_prop"] = [0.5, 0.8, 0.95, 1.0]
    case_1_expect["proportion"] = [0.5, 0.3, 0.15, 0.05]
    pd.testing.assert_frame_equal(case_1, case_1_expect)
    # only some
    case_2 = by_combined_proportions(
        intervals_overshoot=intervals,
        proportion_tshd=1.0,
        proportion_indiv_tshd=0.1,
    )
    case_2_expect = case_1.copy().iloc[0:3]
    pd.testing.assert_frame_equal(case_2, case_2_expect)
    # only one
    case_3 = by_combined_proportions(
        intervals_overshoot=intervals,
        proportion_tshd=0.5,
        proportion_indiv_tshd=0.1,
    )
    case_3_expect = case_1.copy().iloc[0:1]
    pd.testing.assert_frame_equal(case_3, case_3_expect)
    # 0 proportion : always at least one value (same as case 3)
    case_4 = by_combined_proportions(
        intervals_overshoot=intervals,
        proportion_tshd=0.0,
        proportion_indiv_tshd=1.0,
    )
    assert case_4.empty
    # custom reference
    case_5 = by_combined_proportions(
        intervals_overshoot=intervals,
        proportion_tshd=0.0,
        proportion_indiv_tshd=1.0,
        energy_reference=5000.0,
    )
    case_5_expect = case_3_expect.copy()
    case_5_expect["proportion"] = [1.0]
    case_5_expect["cum_energy_prop"] = [1.0]
    pd.testing.assert_frame_equal(case_5, case_5_expect)
    # check not inplace
    with pytest.raises(KeyError):
        intervals["cum_energy_prop"]



"""Tests for the ``MeanProfile`` base class.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.mean_profile import MeanProfile

from .check import compare_profiles
from .fake_data import sinusoid_history


def test_daily_1():
    history = sinusoid_history(freq="30min", noise=0, n_days=7)
    profile = MeanProfile(period="D")
    daily = profile.compute(history, history.index[-1].ceil("D"))
    expected = history.iloc[:48].copy()
    expected.index += 7 * pd.Timedelta("1D")
    compare_profiles(daily, expected)


def test_daily_2():
    history = sinusoid_history(freq="15min", noise=0.05, n_days=30)
    profile = MeanProfile(period="D")
    daily = profile.compute(history, history.index[-1].ceil("D"))
    expected = history.iloc[:96].copy()
    expected.index += 30 * pd.Timedelta("1D")
    assert expected.size == daily.size



"""Tests for the ``HybridThreshold``.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.thresholds.hybrid_rel_std import (
    HybridThreshold,
)

from .check import compare_profiles
from .fake_data import sinusoid_history


def test_daily_1():
    """Check that 0 relative offsets falls back to mean profile"""
    history = sinusoid_history(freq="30min", noise=0, n_days=7)
    profile = HybridThreshold(
        period="D",
        offset_std=0,
        offset_relative=0,
        window=1,
    )
    daily = profile.compute(history, history.index[-1].ceil("D"))
    expected = history.iloc[:48].copy()
    expected.index += 7 * pd.Timedelta("1D")
    compare_profiles(daily, expected)


def test_daily_2():
    """Check with 3 srt relative offset on periodic time-series.

    TODO : Result not verified.
    """
    history = sinusoid_history(freq="30min", noise=0.5, n_days=7)
    profile = HybridThreshold()
    profile.compute(history, history.index[-1].ceil("D"))



"""Generation of fake data to test the library.
"""

import numpy as np
import pandas as pd
import scipy.constants as SK


def sinusoid_history(
    start=None,
    n_days=21,
    freq="30min",
    noise=0.05,
    period_variation=SK.day,
    min_value=0.0,
    max_value=1.0,
):
    """Return a series of fake historical data."""
    duration = pd.Timedelta(days=n_days)
    if start is None:
        start = pd.Timestamp.now().floor("D") - duration
    end = start + duration
    index_dates = pd.date_range(start=start, freq=freq, end=end, inclusive="left")
    # TODO : factorize using cerebro utils (extracted)
    durations = (index_dates - index_dates[0]).total_seconds()
    omega = (2 * np.pi) / period_variation
    data = (np.sin(durations * omega) + 1 + min_value) / 2 * max_value
    data += np.random.uniform(low=-1 * noise, high=noise, size=data.shape[0]) * (
        max_value - min_value
    )
    history = pd.Series(data, index=index_dates, name="example")
    return history


def sinusoid_history_df(*args, **kwargs):
    """Return a series of fake historical data as a DataFrame

    The function wraps :py:func:`sinusoid_history` and transforms the result
    to a DataFrame.
    """
    history_series = sinusoid_history(*args, **kwargs)
    return pd.DataFrame.from_dict({history_series.name: history_series})



"""Defines custom errors about resampling."""
from .base import EATEmptyDataError, EATError


class EATResamplingError(EATError):
    """A resampling operation is impossible.

    This base class is used when an invalid resampling operation is attempted in |eat|.
    Derived classed may be used for more specific resampling errors.
    """


class EATEmptySourceError(EATResamplingError, EATEmptyDataError):
    """Resampling an empty timeseries to specific instants is impossible.

    This exception is used when :

    - a resampling operation is undefined/meaningless when the source data is empty,
    - empty data is passed as source.

    """


class EATEmptyTargetsError(EATResamplingError, EATEmptyDataError):
    """Resampling a timeseries to empty targets is meaningless in this situation.

    This exception is used when :

    - a resampling operation is undefined/meaningless when the target instants
      set is empty,
    - empty data is passed as targets.

    """



""" """

import functools

import numpy as np
import pandas as pd
import scipy.constants as SK

from energy_analysis_toolbox.timeseries.profiles.preprocessing.history_filters.categories import (
    keep_categories,
    remove_categories,
    same_category,
)

from .fake_data import sinusoid_history_df


# =============================================================================
# Utility functions for test purpose
# =============================================================================
def is_working(timestamp, off_days=None):
    """Return True if the day of week of timestamp is in ``off_days``.

    Parameters
    ----------
    timestamp : pd.Timestamp
    off_days : list
        List of integers between 0 and 6 representing, say, the "working days"
        of a company. 0 is monday.
        Default is None wjocj falls back to ``[]``.

    Returns
    -------
    bool:
        True is the day od ``date`` is not in ``off_days``.

    """
    if off_days is None:
        off_days = []
    return timestamp.dayofweek not in off_days


# Name after an italian restaurant which cooks delicious panzerotti
casa_randazzo = functools.partial(is_working, off_days=[0, 1])


# All in one classification
def single_category(x, **kwargs):
    return 1


# =============================================================================
# same_category
# =============================================================================
def test_same_category_identity():
    """Check that the function is identity with only one category."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df
    filtered = same_category(
        history,
        date=monday + pd.Timedelta("1D"),
        classificator=single_category,
    )
    pd.testing.assert_frame_equal(filtered, history)
    # on series
    filtered_s = same_category(
        history.iloc[:, 0],
        date=monday + pd.Timedelta("1D"),
        classificator=single_category,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[:, 0])


def test_same_category_monday_tuesday():
    """Check the results with two categories : monday/tuesday and others."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df
    filtered = same_category(
        history,
        date=monday + pd.Timedelta("1D"),
        classificator=casa_randazzo,
    )
    pd.testing.assert_frame_equal(filtered, history.iloc[[0, 1, 7, 8], :])
    # on series
    filtered_s = same_category(
        history.iloc[:, 0],
        date=monday + pd.Timedelta("1D"),
        classificator=casa_randazzo,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[[0, 1, 7, 8], 0])


def test_same_category_arbitrary():
    """Check the result using an external categorization with a proxy getter."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    categories = pd.Series(
        np.arange(0, 14) % 2,
        index=history.index,
    ).astype("bool")
    categories.loc[monday + pd.Timedelta("15D")] = False

    def classificator(date):
        return categories.loc[date]

    # on df
    filtered = same_category(
        history,
        date=monday + pd.Timedelta("1D"),
        classificator=classificator,
    )
    pd.testing.assert_frame_equal(filtered, history.loc[categories.values[:-1]])
    # on series
    filtered_s = same_category(
        history.iloc[:, 0],
        date=monday + pd.Timedelta("1D"),
        classificator=classificator,
    )
    pd.testing.assert_series_equal(
        filtered_s,
        history.loc[categories.values[:-1]].iloc[:, 0],
    )


# =============================================================================
# Example with a single category
# =============================================================================


def test_keep_categories_constant():
    """Check that the function works as expected with simple case of one category."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df
    filtered = keep_categories(history, keep=[1], classificator=single_category)
    pd.testing.assert_frame_equal(filtered, history)
    filtered_out = keep_categories(history, keep=[0], classificator=single_category)
    assert filtered_out.empty
    # on series
    filtered_s = keep_categories(
        history.iloc[:, 0],
        keep=[1],
        classificator=single_category,
    )
    filtered_out_s = keep_categories(
        history.iloc[:, 0],
        keep=[0],
        classificator=single_category,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[:, 0])
    assert filtered_out_s.empty


def test_remove_categories_constant():
    """Check that the function works as expected with simple case of one category."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df
    filtered = remove_categories(history, remove=[0], classificator=single_category)
    pd.testing.assert_frame_equal(filtered, history)
    filtered_out = remove_categories(history, remove=[1], classificator=single_category)
    assert filtered_out.empty
    # on series
    filtered_s = remove_categories(
        history.iloc[:, 0],
        remove=[0],
        classificator=single_category,
    )
    filtered_out_s = remove_categories(
        history.iloc[:, 0],
        remove=[1],
        classificator=single_category,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[:, 0])
    assert filtered_out_s.empty


# =============================================================================
# Keep only certain categories
# =============================================================================


def test_keep_categories_None_classificator():
    """Check that None classificator returns the whole history"""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=3,
        freq="1D",
        period_variation=7 * SK.day,
    )
    filtered = keep_categories(history, keep=[True], classificator=None)
    pd.testing.assert_frame_equal(filtered, history)
    filtered_2 = keep_categories(history, keep=None, classificator=None)
    pd.testing.assert_frame_equal(filtered_2, history)


def test_keep_categories_working_days():
    """Check with a function which returns True except on monday and tuesday."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df keep nothing
    assert keep_categories(history, classificator=casa_randazzo).empty
    # on df keep everything
    pd.testing.assert_frame_equal(
        keep_categories(history, classificator=casa_randazzo, keep=[True, False]),
        history,
    )
    # keep working days
    filtered = keep_categories(history, keep=[True], classificator=casa_randazzo)
    pd.testing.assert_frame_equal(
        filtered,
        history.drop(labels=history.index[[0, 1, 7, 8]]),
    )
    # on series
    filtered_s = keep_categories(
        history.iloc[:, 0],
        keep=[False],
        classificator=casa_randazzo,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[[0, 1, 7, 8], 0])


def test_keep_categories_arbitrary():
    """Check when getting categories from a proxy table."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    categories = pd.Series(
        ["no", "yes"] * 7,  # no on even indices
        index=history.index,
    )

    def classificator(date):
        return categories.loc[date]

    # on df
    filtered = keep_categories(history, keep=["yes"], classificator=classificator)
    pd.testing.assert_frame_equal(filtered, history.iloc[1::2])
    # on series
    filtered_s = keep_categories(
        history.iloc[:, 0],
        keep=["no"],
        classificator=classificator,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[0::2, 0])


# =============================================================================
# Remove only certain categories
# =============================================================================


def test_remove_categories_None_classificator():
    """Check that None classificator returns the whole history"""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=3,
        freq="1D",
        period_variation=7 * SK.day,
    )
    filtered = remove_categories(history, classificator=None)
    pd.testing.assert_frame_equal(filtered, history)


def test_remove_categories_working_days():
    """Check with a function which returns True except on monday and tuesday."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    # on df remove everything
    assert remove_categories(
        history,
        classificator=casa_randazzo,
        remove=[True, False],
    ).empty
    # on df remove nothing
    pd.testing.assert_frame_equal(
        remove_categories(history, classificator=casa_randazzo, remove=[]),
        history,
    )
    # remove working days
    filtered = remove_categories(history, remove=[False], classificator=casa_randazzo)
    pd.testing.assert_frame_equal(
        filtered,
        history.drop(labels=history.index[[0, 1, 7, 8]]),
    )
    # on series
    filtered_s = remove_categories(
        history.iloc[:, 0],
        remove=[True],
        classificator=casa_randazzo,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[[0, 1, 7, 8], 0])


def test_remove_categories_arbitrary():
    """Check when getting categories from a proxy table."""
    monday = pd.Timestamp("2022-01-03")
    history = sinusoid_history_df(
        start=monday,
        n_days=14,
        freq="1D",
        period_variation=7 * SK.day,
    )
    categories = pd.Series(
        ["no", "yes"] * 7,  # no on even indices
        index=history.index,
    )

    def classificator(date):
        return categories.loc[date]

    # on df
    filtered = remove_categories(history, remove=["no"], classificator=classificator)
    pd.testing.assert_frame_equal(filtered, history.iloc[1::2])
    # on series
    filtered_s = remove_categories(
        history.iloc[:, 0],
        remove=["yes"],
        classificator=classificator,
    )
    pd.testing.assert_series_equal(filtered_s, history.iloc[0::2, 0])



"""Provide a comprehensive suite of tools for energy data analysis.

This module is designed to help users analyze energy consumption data using a variety
of methods, including thermosensitivity modeling, weather data integration, time series
analysis, synthetic data generation, and more.

Main Features
-------------
- **Energy Data Analysis**: Tools to process, transform, and analyze energy consumption
  data from buildings or systems.
- **Thermosensitivity Modeling**: Classes for modeling energy consumption sensitivity
  to temperature, including degree days computations.
- **Categorical Analysis**: Functions to categorize energy consumption data by
  time-based metrics, such as days of the week or custom time intervals.
- **Weather Integration**: Tools for working with weather data to analyze the influence
  of external conditions on energy consumption.
- **Synthetic Data Generation**: Methods for creating synthetic datasets that simulate
  thermosensitive energy consumption patterns.
- **Time Series Profiling**: Features for creating, analyzing, and categorizing energy
  time series profiles to derive meaningful insights.
- **Logger and Utilities**: A set of logging tools for detailed traceability and
  utility functions for everyday energy data operations.

Submodules
----------
- **constants**: Contains shared constants used throughout the module for energy and
  temperature analysis.
- **pandas**: Utilities for working with Pandas dataframes specific to energy data
  processing.
- **logger**: A logging configuration and utility to enable detailed logging and
  diagnostics for energy analysis routines.
- **keywords**: Defines specific keywords used throughout the analysis for consistency.
- **weather**: Includes tools to calculate heating and cooling degree days and methods
  to analyze thermosensitivity against weather conditions.
- **timeseries**: Modules for time series analysis, including profiles for
  localization, rolling means, thresholds, and other profiling methods.
- **synthetic**: Provides utilities for generating synthetic energy consumption
  datasets with thermosensitivity properties.
- **tests**: Includes unit tests for verifying the correctness of different analysis
  tools and ensuring the module's integrity.

Examples
--------
To begin analyzing energy data:

>>> from energy_analysis_toolbox.weather import degree_days
>>> from energy_analysis_toolbox.timeseries.profiles import mean_profile
>>> import pandas as pd
>>> temperature_data = pd.Series([...], index=pd.date_range(start="2023-01-01",
... periods=365, freq="D"))
>>> degree_days_data = degree_days.dd_compute(temperature_data, base_temperature=18)
>>> mean_profile_result = mean_profile.calculate_mean_profile(temperature_data)

This example demonstrates how to compute degree days and create a mean temperature
profile using the module's submodules.

Notes
-----
This module is intended for energy data analysts who need to assess consumption trends,
model thermosensitivity, and understand how weather impacts energy use. The toolbox
contains both statistical tools for detailed analysis and utility functions for
preprocessing and structuring energy data.

"""

from . import (
    constants,
    energy,
    errors,
    keywords,
    power,
    synthetic,
    tests,
    thermosensitivity,
    timeseries,
)

__version__ = "0.1.0"



"""Functions to transform and merge overconsumption intervals after they are located.

Provides functions for transforming overconsumption intervals after
they have been located, including merging nearby intervals and recalculating
their duration and energy contributions.

"""

import pandas as pd

from energy_analysis_toolbox.timeseries.create.from_intervals import flatten_and_fill
from energy_analysis_toolbox.timeseries.extract_features import (
    intervals_over,
    timestep_durations,
)


def merge_by_proximity(
    intervals_overshoot: pd.DataFrame,
    min_interval: float = 600,
) -> pd.DataFrame:
    """Return a table where the overconsumption events too short are merged.

    Parameters
    ----------
    intervals_overshoot : pd.DataFrame
        A table of overshoot overconsumption with at least 'start', 'end' and 'energy'
        columns.
    min_interval : float, optional
        The minimum duration in (s) to be imposed between two overshoot overconsumption.
        All overconsumption separated by a duration under this threshold are merged.
        Default is 600 seconds corresponding to 10 minutes.

    Returns
    -------
    overconsumption : pd.DataFrame
        A table of overconsumption with 'start', 'end' and 'energy' obtained after
        merging the time-neighboring overconsumption in ``intervals_overshoot``.


    .. note::

        The energy of the result of the merge between two overconsumption is the sum
        of the interval energies.

    Notes
    -----
    The function proceeds as follows :

    - [1] flatten the overconsumption to a table of timeseries for each variable. Fill
      all values between overconsumption with zeros.
    - [2] Recompute the durations so that the "duration" variable has the right
      value between the overshoot overconsumption.
    - [3] Drop all the rows for which the duration is under the defined threshold
      and the energy is 0. By construction, overshoot overconsumption have a non-zero
      energy (they are overshoots!) while other rows were filled with 0. Overshoots
      which were closed than the threshold are now contiguous in the timeseries.
    - [4] Re-extract the periods during which the energy is > 0 : the contiguous
      overconsumption are merged by this process.
    - [5] Recompute the duration and energy for the new overconsumption.

    Some limit cases are managed :

    - empty data is returned directly
    - contiguous overconsumption (which are a limit limit case) are managed by dropping
      the "interstitial" row with duplicate index which appears in the flattened
      overconsumption. This case should not be encountered.

    """
    if intervals_overshoot.empty:
        return intervals_overshoot
    # 1
    flat_intervals = flatten_and_fill(intervals_overshoot.sort_index()).fillna(0)
    # 2
    flat_intervals["duration"] = timestep_durations(
        flat_intervals["duration"],
        last_step=flat_intervals["duration"].iloc[-1],
    )
    # Adjacent overconsumption create duplicates
    flat_intervals = flat_intervals.loc[~flat_intervals.index.duplicated(keep="first")]
    # 3
    dropped = flat_intervals.iloc[:-1, :].query(
        f"energy == 0 and duration < {min_interval}",
    )
    flat_intervals = flat_intervals.drop(index=dropped.index)
    flat_intervals["duration"] = timestep_durations(flat_intervals["duration"])
    # 4
    intervals = intervals_over(flat_intervals["energy"], 0.0)
    # 5
    intervals["duration"] = (intervals["end"] - intervals["start"]).dt.total_seconds()
    # 5 inclusive loc but energy filled with 0 between the overshoots.
    intervals["energy"] = intervals.apply(
        lambda x: (flat_intervals.loc[x["start"] : x["end"], "energy"].sum()),
        axis=1,
    )
    return intervals



"""Provide tools for integrating and analyzing weather data in energy analysis.

This package is designed to facilitate the computation of degree days and analyze the
influence of outdoor temperature on energy consumption. It includes methods to
calculate different types of degree days (e.g., heating and cooling).

Main Features
-------------
- **Degree Days Computation**: Functions to compute heating, cooling, and both types
  of degree days to measure energy needs based on temperature.
- **Flexible Calculation Methods**: Supports multiple methods for calculating degree
  days, such as integral, mean temperature, and min-max methods.
- **Temperature Analysis**: Tools to understand and model the effect of temperature
  on energy consumption, helping to quantify the impact of weather conditions.

"""



r"""Process the thermosensitivity data.

# Available classes

## ThermoSensitivity

Class to compute the thermosensitivity of a building.
Needs a time series of energy consumption and outdoor temperature.

## CategoricalThermoSensitivity

Class to compute the thermosensitivity of a building with labeled periods.
Needs a time series of energy consumption, outdoor temperature, and labeled periods.

The labeled periods are resampled to the same frequency as the energy and temperature
data by taking the most common category in the period.

Currently, the class only calibrates one base temperature for all the categories
aggregated.

# Implementation details

## Resampling frequency

The energy and temperature data are resampled at a given frequency.
The degree days are computed at the same frequency.

## Thermo Sensitivity

The thermo-sensitivity is modelled as a linear regression between the energy
consumption and the degree days.

    .. math::

        E ~ E0 + TS \times DegreeDays

The degree days are computed from the temperature data and the base temperature.

    .. math::

        DegreeDays = \\int max(0, BaseTemperature - T(t)) dt

Different methods are available to compute the degree days:

- Integral: sum the difference between the base temperature and the temperature.
    .. math::
        DegreeDays = \\sum_{t=0}^N max(0, BaseTemperature - T(t))

- Mean: sum the difference between the base temperature and the mean temperature.
    .. math::
        DegreeDays = max(0, BaseTemperature - \\bar{T} )

- MinMax: sum the difference between the base temperature and the mean temperature
  computed as the mean of the minimum and maximum temperature.

    .. math::
        DegreeDays = max(0, BaseTemperature - \\frac{T_{min} + T_{max}}{2} )

See the `dd_compute` function in the `energy_analysis_toolbox.weather.degree_days`
module.

Over a long period, the data can present a thermosensitivity with different types of
degree days:

- Heating: the energy consumption increases when the temperature decreases. Usually,
  the base temperature is around 18°C.
- Cooling: the energy consumption increases when the temperature increases. Usually,
  the base temperature is around 24°C.

## Auto-calibration

Two aspects of the thermosensitivity can be automatically detected:

- The degree days type: heating, cooling, or both.
- The base temperature.

### Degree days type

Each building, depending of the installed systems, can have different thermosensitivity
types :
- use of heating systems (heating degree days, during the winter)
- use of cooling systems (cooling degree days, during the summer)
- use of both systems (heating and cooling degree days)

The degree days type can be automatically detected by setting the `degree_days_type`
parameter to ``"auto"``. The method will compute the Spearman correlation between the
energy and the temperature for the periods with the mean temperature below and above
the intersaison mean temperature (default is 20°C).

### Base temperature

The base temperature can be calibrated by minimizing the mean squared error between the
data and the model.

Each degree days type has a specific base temperature that is determined by analyzing
the data over the corresponding periods. The heating (resp. cooling) base temperature
is calibrated by minimizing the mean squared error between the energy and the heating
(resp. cooling) degree days for the periods with the mean temperature below (resp.
above) the intersaison mean temperature.

The optimization is done with the `scipy.optimize.minimize_scalar` function with the
`bounded` method.

"""

import logging
from functools import cached_property
from typing import TYPE_CHECKING, Literal, TypeVar, cast

import numpy as np
import pandas as pd
import statsmodels
from scipy.optimize import minimize_scalar
from scipy.stats import spearmanr
from statsmodels.api import OLS

from energy_analysis_toolbox.energy.resample import to_freq as energy_to_freq
from energy_analysis_toolbox.logger import init_logging
from energy_analysis_toolbox.weather.degree_days import (
    dd_compute,
    dd_types,
    literal_computation_dd_types,
    literal_dd_types,
    literal_valid_dd_types,
)

if TYPE_CHECKING:
    from pandas.core.resample import Resampler


ThermosensitivityInstance = TypeVar(
    "ThermosensitivityInstance",
    bound="ThermoSensitivity",
)


class ThermoSensitivity:
    """Class to compute the thermosensitivity of a building.

    Examples
    --------
    >>> from energy_analysis_toolbox.thermosensitivity import ThermoSensitivity
    >>> import pandas as pd
    >>> import numpy as np
    >>> np.random.seed(0)
    >>> temperature_data = 15 + 2*pd.Series(np.random.rand(366),
    ... index=pd.date_range("2020-01-01", "2020-12-31", freq="D"))
    >>> energy_data = 10 + (16 - temperature_data).clip(0) * 5 + np.random.rand(366)
    >>> ts = ThermoSensitivity(energy_data, temperature_data, degree_days_type="auto")
    >>> ts.fit()
    >>> ts
    ThermoSensitivity(frequency=1D,
            degree_days_type=heating,
            degree_days_base_temperature={"heating": 15.98},
            degree_days_computation_method="integral",
            interseason_mean_temperature=20)
    <BLANKLINE>
                                OLS Regression Results
    ==============================================================================
    Dep. Variable:                 energy   R-squared:                       0.969
    Model:                            OLS   Adj. R-squared:                  0.969
    No. Observations:                 366   F-statistic:                 1.137e+04
    Covariance Type:            nonrobust   Prob (F-statistic):          1.25e-276
    =======================================================================================
                              coef    std err          t      P>|t|      [0.025   \
                                  0.975]
    ---------------------------------------------------------------------------------------
    heating_degree_days     5.1177      0.048    106.638      0.000       5.023    \
        5.212
    Intercept              10.5120      0.019    539.733      0.000      10.474    \
        10.550
    =======================================================================================
    <BLANKLINE>
    Notes:
    [1] Standard Errors assume that the covariance matrix of the errors is correctly
    specified.

    As you can see from the example above:

    - the type of thermosensitivity is automatically detected (heating in this case)
    - the base temperature is calibrated to 15.98 (true value is 16)
    - the model is fitted with a R-squared of 0.969
    - the heating degree days coefficient is 5.1177 (true value is 5)
    - the intercept is 10.5120 (true value is 10)

    """

    target_name = "energy"
    temperature_name = "temperature"

    def __init__(
        self,
        energy_data: pd.Series,
        temperature_data: pd.Series,
        frequency: str = "1D",
        degree_days_type: literal_dd_types = "heating",
        degree_days_base_temperature: dict | None = None,
        degree_days_computation_method: literal_computation_dd_types = "integral",
        interseason_mean_temperature: float = 20,
        base_logger_name: str | None = None,
        min_logger_level_stdout: int | str = logging.ERROR,
    ) -> None:
        """Initialize a ``ThermoSensitivity`` instance.

        Parameters
        ----------
        energy_data : pd.Series
            Time series of energy consumption data for the building.
        temperature_data : pd.Series
            Time series of outdoor temperature data.
        frequency : str, optional
            Frequency for resampling the data (default is "1D").
            Options include:
            - "1D": daily resampling
            - "7D": weekly resampling
            - "1W-MON": weekly resampling starting on Monday.
        degree_days_type : str, optional
            Type of degree days to compute (default is "heating").
            Options are:
            - "heating": heating degree days.
            - "cooling": cooling degree days.
            - "both": both heating and cooling degree days.
            - "auto": automatically detect the degree days type.
        degree_days_base_temperature : dict, optional
            Base temperature(s) for degree day calculations (default is an empty dict).
            Should include keys "heating" and/or "cooling".
        degree_days_computation_method : str, optional
            Method to compute degree days (default is "integral").
            Options are:
            - "integral": integral calculation.
            - "mean": mean temperature calculation.
            - "min_max": min-max temperature calculation.
            - "pro": energy-professionals calculation.
        interseason_mean_temperature : float, optional
            Mean temperature to differentiate heating and cooling periods (default 20).
        base_logger_name : str, optional
            Name of the logger. By default, it is the class name. All following
            following instances receive a unique identifier, based on the first
            one, with the pattern:

            - ``<base_logger_name>``
            - ``<base_logger_name>_1``
            - ...

        min_logger_level_stdout: str, int, optional
            Minimum logger level below which no message is transferred to stdout
            (i.e. not printed). Default is ``"ERROR"``.

        Raises
        ------
        ValueError
            If the degree days type is not valid or if required parameters are missing.

        Example
        -------
        >>> ts = ThermoSensitivity(energy_data, temperature_data,
            degree_days_type="auto")

        """
        self._init_logger(
            base_logger_name=base_logger_name,
            min_logger_level_stdout=min_logger_level_stdout,
        )
        self._energy_data = energy_data
        self._temperature_data = temperature_data
        self._frequency = frequency
        self._aggregated_data: pd.DataFrame | None = None
        self.degree_days_type = degree_days_type
        self.degree_days_base_temperature = degree_days_base_temperature or {}
        self.degree_days_computation_method = degree_days_computation_method
        self.interseason_mean_temperature = interseason_mean_temperature
        self.predictors: list[str] = []
        self._model = None
        self._validate_data()
        self._post_init()

    def _init_logger(
        self,
        base_logger_name: str | None,
        min_logger_level_stdout: str | int = logging.ERROR,
    ) -> None:
        """Initialize the logger for the current instance of the class.

        This method configures a logger specific to the instance of the class. If a
        `base_logger_name` is provided, it will be used as the logger's name; otherwise,
        the class name will be used. The logger is set up to manage output levels,
        allowing only messages with a severity level above the specified minimum to be
        printed to standard output.

        Parameters
        ----------
        base_logger_name : str, optional
            The name of the logger. If not provided, the logger will default to the
            name of the class. Subsequent instances will receive unique logger names
            by appending an incrementing counter to the base name.

        min_logger_level_stdout : str or int, optional
            The minimum severity level for messages to be printed to standard output.
            Messages below this level will not be displayed. The default is
            ``"ERROR"``, meaning only errors and critical messages will
            be printed. Acceptable values can be specified as strings (e.g.,
            "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL") or as corresponding
            integer constants from the `logging` module. If you want to have details
            on what's going on in the code, set this to ``"INFO"`` or ``"WARNING"``.

        Returns
        -------
        None
            This method does not return a value; it initializes the logger instance
            for use in logging messages throughout the class.

        Example
        -------
        >>> ts = ThermoSensitivity(energy_data, temperature_data)
        >>> ts._init_logger(base_logger_name="my_custom_logger",
        ... min_logger_level_stdout=logging.ERROR)

        Notes
        -----
        - Ensuring unique logger names prevents conflicts when multiple instances of
          the class are used simultaneously.
        - Log messages can be directed to different outputs (e.g., console, file) based
          on logger configuration.

        """
        if not base_logger_name:
            base_logger_name = self.__class__.__name__
        self.logger = init_logging(
            base_logger_name=base_logger_name,
            min_logger_level_stdout=min_logger_level_stdout,
        )

    @property
    def frequency(
        self,
    ) -> str:
        """The frequency of the resampled data.

        The property is unmutable. To change the frequency, create a new object.
        """
        return self._frequency

    @property
    def energy_data(
        self,
    ) -> pd.Series:
        """The energy data of the building.

        The property is unmutable. To change the energy data, create a new object.
        """
        return self._energy_data

    @property
    def temperature_data(
        self,
    ) -> pd.Series:
        """The outdoor temperature data.

        The property is unmutable. To change the temperature data, create a new object.
        """
        return self._temperature_data

    @cached_property
    def resampled_energy(
        self,
    ) -> pd.Series:
        """The energy data resampled at the given frequency.

        Uses the `to_freq` function from the `energy_analysis_toolbox.energy.resample`
        module to convert the energy data to the desired frequency.

        This property is cached to avoid recomputing it multiple times.
        """
        energy = self.energy_data.copy()
        last_period = energy.index[-1] - energy.index[-2]
        energy[energy.index[-1] + last_period] = 0
        return energy_to_freq(
            energy,
            self.frequency,
        ).rename(self.target_name)

    @cached_property
    def resampled_temperature(
        self,
    ) -> pd.Series:
        """The temperature data resampled at the given frequency.

        Average the temperature data over the given frequency.

        This property is cached to avoid recomputing it multiple times.
        """
        return (
            self.temperature_data.resample(self.frequency)
            .mean()
            .rename(self.temperature_name)
        )

    @cached_property
    def resampled_energy_temperature(
        self,
    ) -> pd.DataFrame:
        """The resampled energy and temperature data.

        The DataFrame contains the resampled energy and temperature data.
        Periods with missing values are removed.
        """
        return pd.concat(
            [self.resampled_energy, self.resampled_temperature],
            axis=1,
        ).dropna(how="any", axis=0)

    @property
    def model(
        self,
    ) -> statsmodels.regression.linear_model.RegressionResults:
        """The thermosensitivity model.

        A `statsmodels.regression.linear_model.RegressionResults` object.

        Raises
        ------
        ValueError
            If the model is not fitted. Use the `fit` method to train the model.

        """
        if self._model is None:
            err = "Model not fitted. Please run the `fit` method."
            raise ValueError(err)
        return self._model

    @property
    def aggregated_data(
        self,
    ) -> pd.DataFrame:
        """The aggregated data used to fit the model.

        The Data is a DataFrame resampled at the provided Frequency with the following
        columns:
        - "energy": the total energy at the frequency.
        - "temperature": the mean temperature at the frequency.
        - (Optional) "heating_degree_days": the heating degree days at the frequency.
        - (Optional) "cooling_degree_days": the cooling degree days at the frequency.

        Raises
        ------
        ValueError
            If the data is not aggregated. Use the `fit` method to aggregate the data.

        """
        if self._aggregated_data is None:
            err = "Data not aggregated. Please run the `fit` method."
            raise ValueError(err)
        return self._aggregated_data

    @aggregated_data.setter
    def aggregated_data(
        self,
        value: pd.DataFrame,
    ) -> None:
        """Set the aggregated data, and reset the model."""
        self._aggregated_data = value
        self._model = None

    def _validate_data(
        self,
    ) -> None:
        """Check the validity of the parameters.

        Raises
        ------
        ValueError
            If the degree days type is not valid.
        ValueError
            If the base temperature is not specified for the heating or cooling degree
            days. While not empty.

        """
        if self.degree_days_type not in dd_types:
            err = f"Invalid degree days type. Must be one of {dd_types}."
            raise ValueError(err)
        if self.degree_days_base_temperature != {}:
            if self.degree_days_type in ["heating", "both"]:
                try:
                    self.degree_days_base_temperature["heating"]
                except KeyError:
                    err = (
                        "Base temperature for heating degree days must be specified."
                        "\n Example: degree_days_base_temperature={'heating': 18, "
                        "'cooling': 24}"
                    )
                    raise ValueError(err) from None
            elif self.degree_days_type in ["cooling", "both"]:
                try:
                    self.degree_days_base_temperature["cooling"]
                except KeyError:
                    err = (
                        "Base temperature for cooling degree days must be specified."
                        "\n Example: degree_days_base_temperature={'heating': 18, "
                        "'cooling': 24}"
                    )
                    raise ValueError(err) from None

    def _post_init(
        self,
    ) -> None:
        """End the initialization process.

        If the degree days type is set to ``"auto"``, the method will detect the degree
        days type. See :meth:`_detect_degree_days_type`. After the detection, the
        predictors will be set.
        """
        self._detect_degree_days_type()
        if self.degree_days_type == "both":
            self.predictors = ["heating_degree_days", "cooling_degree_days"]
        elif self.degree_days_type in ["heating", "cooling"]:
            self.predictors = [f"{self.degree_days_type}_degree_days"]

    def _aggregate_data(
        self,
        degree_days_base_temperature: dict,
    ) -> None:
        """Compute the degree days and aggregate the data.

        Store the aggregated data in the `aggregated_data` property.
        """
        degree_days = self._calculate_degree_days(degree_days_base_temperature)
        self.aggregated_data = pd.concat(
            [
                self.resampled_energy_temperature,
                degree_days,
            ],
            axis=1,
        )

    def _detect_degree_days_type(
        self,
        significance_level: float = 0.05,
    ) -> None:
        """Estimate the degree days type if it is set to ``"auto"``.

        It will compute the Spearman correlation (with the alternative hypothesis)
        between the energy and the temperature. If the p-value is below the
        significance level, the degree days type will be set to ``"heating"`` or
        ``"cooling"`` or ``"both"``.

        If no significant correlation is found, the method will raise a ValueError.

        - **Heating**: The energy consumption is negatively correlated with the
          temperature for the periods with the mean temperature below the intersaison
          mean temperature.
        - **Cooling**: The energy consumption is positively correlated with the
          temperature for the periods with the mean temperature above the intersaison
          mean temperature.

        Note:
        ----
        The Spearman correlation is a non-parametric test that measures the strength
        and direction of the monotonic relationship between two variables. The relation
        is not necessarily linear, and the test does not assume that the data is
        normally distributed.

        .. warning::
            If the data contains multiple groups, the Simpson's paradox may occur. The
            Paradox states that the correlation observed in the aggregated data may not
            hold when the data is split into subgroups. See the
            `CategoricalThermoSensitivity` class for a solution.

            References: `wikipedia <https://en.wikipedia.org/wiki/Simpson%27s_paradox>`_

        """
        if self.degree_days_type == "auto":
            heating_mask = (
                self.resampled_energy_temperature[self.temperature_name]
                < self.interseason_mean_temperature
            )
            min_nb_points_heating = 10
            if sum(heating_mask) <= min_nb_points_heating:
                warn = (
                    "Not enough data for the heating period."
                    f"Number of data points: {sum(heating_mask)}"
                )
                self.logger.warning(warn)
                # too few point to do any test
                heating_sp = 1
            else:
                heating_sp = spearmanr(
                    self.resampled_energy_temperature.loc[
                        heating_mask,
                        [self.target_name, self.temperature_name],
                    ],
                    alternative="less",
                ).pvalue
            cooling_mask = (
                self.resampled_energy_temperature[self.temperature_name]
                > self.interseason_mean_temperature
            )
            min_nb_points_cooling = 10
            if sum(cooling_mask) <= min_nb_points_cooling:
                warn = (
                    "Not enough data for the cooling period."
                    f"Number of data points: {sum(heating_mask)}"
                )
                self.logger.warning(warn)
                # too few point to do any test
                cooling_sp = 1
            else:
                data_to_test = self.resampled_energy_temperature.loc[
                    cooling_mask,
                    [self.target_name, self.temperature_name],
                ]
                cooling_sp = spearmanr(data_to_test, alternative="greater").pvalue
            if heating_sp < significance_level and cooling_sp < significance_level:
                self.degree_days_type = "both"
            elif heating_sp < significance_level:
                self.degree_days_type = "heating"
            elif cooling_sp < significance_level:
                self.degree_days_type = "cooling"
            else:
                err = (
                    "Cannot detect the degree days type. Please specify it manually.\n"
                    f"{cooling_sp=} \n"
                    f"{heating_sp=} \n"
                    f"{self.resampled_energy_temperature=} \n"
                    f"{self.resampled_energy=} \n"
                    f"{self.resampled_temperature=} \n"
                )
                raise ValueError(err)

    def _calculate_degree_days(
        self,
        degree_days_base_temperature: dict,
    ) -> pd.DataFrame:
        """Compute the degree days.

        Parameters
        ----------
        degree_days_base_temperature : dict
            Base temperature for the computation of the degree days.
            Must be a dictionary with the keys ``"heating"`` and/or ``"cooling"``.
            Example: degree_days_base_temperature={'heating': 18, 'cooling': 24}

        Returns
        -------
        pd.DataFrame
            DataFrame with the heating and/or cooling degree days
            sampled at the given frequency.

        """
        for dd_type in degree_days_base_temperature:
            if self.degree_days_type in [dd_type, "both"]:
                degree_days = [
                    dd_compute(
                        self.temperature_data,
                        degree_days_base_temperature[dd_type],
                        dd_type=dd_type,
                        method=self.degree_days_computation_method,
                    )
                    .resample(self.frequency)
                    .sum()
                    for dd_type in degree_days_base_temperature
                ]
        return pd.concat(degree_days, axis=1)

    def calibrate_base_temperature(
        self,
        dd_type: literal_dd_types = "heating",
        t0: float | None = None,
        xatol: float = 1e-1,
    ) -> float:
        """Calibrate the base temperature for the specified degree days type.

        This method optimizes the base temperature used to compute the degree days
        by minimizing the mean squared error between the energy consumption data and
        the degree days model. The optimization is done using the
        `scipy.optimize.minimize_scalar` function with a bounded method.

        Parameters
        ----------
        dd_type : str, optional
            The type of degree days to calibrate, must be one of the following:
            - "heating": to calibrate the base temperature for heating degree days.
            - "cooling": to calibrate the base temperature for cooling degree days.
            The default is "heating".

        t0 : float, optional
            The initial guess for the base temperature.
            If not provided, the default initial guess is 16°C for heating or 24°C for
            cooling.

        xatol : float, optional
            The absolute error tolerance for the optimization.
            This controls how precise the optimized base temperature needs to be.
            Default is 1e-1 (0.1°C).

        Returns
        -------
        float
            The optimized base temperature for the specified degree days type.

        Raises
        ------
        ValueError
            If the `dd_type` is invalid (not one of "heating" or "cooling").

        Example
        -------
        >>> ts.calibrate_base_temperature(dd_type="heating", t0=15, xatol=0.05)

        """
        if dd_type not in ["heating", "cooling"]:
            err = "Invalid degree days type. Must be one of 'heating' or 'cooling'."
            raise ValueError(err)
        if t0 is None:
            t0 = 16 if dd_type == "heating" else 24
        if dd_type == "heating":
            mask = (
                self.resampled_energy_temperature[self.temperature_name]
                < self.interseason_mean_temperature
            )
            bounds = (10.0, self.interseason_mean_temperature)
        elif dd_type == "cooling":
            mask = (
                self.resampled_energy_temperature[self.temperature_name]
                > self.interseason_mean_temperature
            )
            bounds = (self.interseason_mean_temperature, 30.0)
        else:
            err = "Invalid degree days type. Must be one of 'heating' or 'cooling'."
            raise ValueError(err)
        res = minimize_scalar(
            self.loss_function,
            args=(
                dd_type,
                self.resampled_energy_temperature[self.target_name],
                self.temperature_data,
                self.frequency,
                mask,
                self.degree_days_computation_method,
            ),
            bounds=bounds,
            method="bounded",
            options={
                "xatol": xatol,
            },
        )
        return res.x

    def calibrate_base_temperatures(
        self,
        t0_heating: float | None = None,
        t0_cooling: float | None = None,
        xatol: float = 1e-1,
    ) -> None:
        """Calibrate the base temperatures for both heating and cooling degree days.

        This method optimizes the base temperatures for heating and/or cooling degree
        days by minimizing the mean squared error between the energy consumption data
        and the degree days model. The method will calibrate the base temperatures based
        on the detected or specified `degree_days_type`.

        If the `degree_days_type` is "heating", only the heating base temperature is
        calibrated. If it is "cooling", only the cooling base temperature is calibrated.
        If it is "both", both base temperatures are calibrated.

        Parameters
        ----------
        t0_heating : float, optional
            The initial guess for the heating base temperature. If not provided, the
            default
            initial guess is 16°C.

        t0_cooling : float, optional
            The initial guess for the cooling base temperature. If not provided, the
            default initial guess is 24°C.

        xatol : float, optional
            The absolute error tolerance for the optimization.
            This controls how precise the optimized base temperatures need to be.
            Default is 1e-1 (0.1°C).

        Returns
        -------
        None
            The method updates the `degree_days_base_temperature` attribute with the
            optimized base temperatures for heating and/or cooling.

        Raises
        ------
        ValueError
            If the `degree_days_type` is invalid.

        Example
        -------
        >>> ts.calibrate_base_temperatures(t0_heating=15, t0_cooling=25, xatol=0.05)

        """
        types_to_calibrate = []
        if self.degree_days_type in ["heating", "both"]:
            types_to_calibrate.append("heating")
        if self.degree_days_type in ["cooling", "both"]:
            types_to_calibrate.append("cooling")
        for dd_type in types_to_calibrate:
            if dd_type == "heating":
                t0 = t0_heating
            elif dd_type == "cooling":
                t0 = t0_cooling
            topt = self.calibrate_base_temperature(
                dd_type=cast(Literal["heating", "cooling"], dd_type),
                t0=t0,
                xatol=xatol,
            )
            self.degree_days_base_temperature[dd_type] = topt

    def _fit_thermosensitivity(
        self,
    ) -> None:
        """Fit the thermosensitivity model with interactions for categories.

        This method performs the following steps:

        1. Drops rows with missing values from the aggregated data.
        2. Creates a regression model where the dependent variable is energy consumption
          and the independent variables are the degree days (heating, cooling, or both)
          and category interactions.
        3. Encodes the category labels using one-hot encoding.
        4. Constructs interaction terms between the degree days and the one-hot encoded
          categories.
        5. Fits the Ordinary Least Squares (OLS) model to the data with the degree days
          and interaction terms.

        The fitted model is stored in the ``_model`` attribute.

        """
        data = self.aggregated_data.dropna(how="any", axis=0)
        y = data[self.target_name].copy()
        x = data[self.predictors].copy()
        x["Intercept"] = 1  # add constant
        self._model = OLS(y, x).fit()

    def fit(
        self: ThermosensitivityInstance,
    ) -> ThermosensitivityInstance:
        """Train the model.

        This method will:

        1. Calibrate the base temperature if it is not set.
           See :meth:`calibrate_base_temperature`.
        2. Aggregate the data. This consists of resampling the energy and temperature
           data and the computation of the degree days. See :meth:`_aggregate_data`.
        3. Fit the thermosensitivity model.
           See :meth:`_fit_thermosensitivity`.

        """
        self.calibrate_base_temperatures()
        self._aggregate_data(self.degree_days_base_temperature)
        self._fit_thermosensitivity()
        return self

    def __repr__(
        self,
    ) -> str:
        """Return the representation of the object."""
        class_name = self.__class__.__name__
        header = f"""{class_name}(frequency={self.frequency},
        degree_days_type={self.degree_days_type},
        degree_days_base_temperature={ {k:round(v, 2) for k,v
            in self.degree_days_base_temperature.items()} },
        degree_days_computation_method={self.degree_days_computation_method},
        interseason_mean_temperature={self.interseason_mean_temperature})"""
        if self._model is not None:
            message = f"{header}\n\n{self.model.summary(slim=True)}"
        else:
            message = header
        return message

    def loss_function(
        self,
        t0: float,
        dd_type: literal_valid_dd_types,
        resampled_energy: pd.Series,
        raw_temperature: pd.Series,
        frequency: str,
        mask: pd.Series | None = None,
        degree_days_computation_method: literal_computation_dd_types = "integral",
    ) -> float:
        """Loss function for the optimization of the base temperature.

        Compute the mean squared error (MSE) between the observed energy data and
        the energy model based on degree days. This function is used for calibrating the
        base temperature in the thermosensitivity model by finding the base temperature
        that minimizes this error.

        The degree days are calculated based on the input temperature data, resampled
        to the desired frequency, and compared against the resampled energy data. The
        MSE is used as the objective function to optimize the base temperature.

        Parameters
        ----------
        t0 : float
            Base temperature used to compute degree days.
        dd_type : literal_dd_types
            Type of degree days to compute. Must be one of:
            - "heating": degree days when temperatures are below the base temperature.
            - "cooling": degree days when temperatures are above the base temperature.
        resampled_energy : pd.Series
            The resampled time series of energy data to be modeled.
        raw_temperature : pd.Series
            The original time series of temperature data, which is used to compute
            degree days based on the base temperature.
        frequency : str
            The frequency to resample the data (e.g., "1D" for daily, "7D" for weekly).
        mask : pd.Series or None, optional
            A boolean mask to filter the data before fitting the model. This allows for
            focusing on specific periods (e.g., only heating or cooling periods).
            Default is None.
        degree_days_computation_method : literal_computation_dd_types, optional
            The method used to compute degree days. Must be one of:
            - "integral": sum the difference between the base temperature and actual
            temperature.
            - "mean": use the mean temperature for degree day calculations.
            - "min_max": use the average of daily minimum and maximum temperatures.
            Default is "integral".

        Returns
        -------
        float
            The mean squared error (MSE) of the residuals between the observed
            energy data and the modeled energy data based on degree days.

        Example
        -------
        >>> loss = loss_function(
                t0=18.0,
                dd_type="heating",
                resampled_energy=energy_data,
                raw_temperature=temperature_data,
                frequency="1D",
                degree_days_computation_method="integral",
            )
        >>> print(f"Calculated loss: {loss}")

        """
        degree_days = dd_compute(
            temperature=raw_temperature,
            reference=t0,
            dd_type=dd_type,
            method=degree_days_computation_method,
        )
        degree_days_resampled = (
            degree_days.resample(frequency).sum().rename("degree_days")
        )
        data = pd.concat([resampled_energy, degree_days_resampled], axis=1).dropna(
            how="any",
            axis=0,
        )
        data["Intercept"] = 1
        if mask is not None:
            data = data[mask]
        model = OLS(
            data[resampled_energy.name],
            data[["degree_days", "Intercept"]],
        ).fit()
        log_message = f"{t0=:.4f}, {model.mse_resid:.2f}, {model.mse_total:.2f}"
        self.logger.info(log_message)
        return model.mse_resid


class CategoricalThermoSensitivity(
    ThermoSensitivity,
):
    """Class to compute the thermosensitivity of a building with labeled periods.

    Based on the `ThermoSensitivity` class.

    """

    categories_name = "category"

    def __init__(
        self,
        energy_data: pd.Series,
        temperature_data: pd.Series,
        categories: pd.Series,
        frequency: str = "1D",
        degree_days_type: literal_dd_types = "heating",
        degree_days_base_temperature: dict | None = None,
        degree_days_computation_method: literal_computation_dd_types = "integral",
        interseason_mean_temperature: float = 20,
        base_logger_name: str | None = None,
        min_logger_level_stdout: int | str = logging.ERROR,
    ) -> None:
        """Return a ``CategoricalThermoSensitivity`` instance.

        Parameters
        ----------
        energy_data : pd.Series
            Energy data of the building.
        temperature_data : pd.Series
            Outdoor temperature data.
        categories : pd.Series
            Labels for the periods.
        frequency : str, optional
            Frequency for the analysis the data, by default "1D".
            If "1D", the data will be resampled daily.
            If "7D", the data will be resampled weekly.
            If "1W-MON", the data will be resampled weekly starting on Monday.
        degree_days_type : str, optional
            Type of degree days to compute. Must be one of the following:

            - "heating": compute only heating degree days (temperature below a
              threshold).
            - "cooling": compute only cooling degree days (temperature above a
              threshold).
            - "both": compute both heating and cooling degree days.
            - "auto": automatically detect the degree days type. See the
              `detect_degree_days_type` method.

        degree_days_base_temperature : dict, optional
            Base temperature for the computation of the degree days, by default {}.
            Must be a dictionary with the keys ``"heating"`` and/or ``"cooling"``.
            Example: degree_days_base_temperature={'heating': 18, 'cooling': 24}
        degree_days_computation_method : str, optional
            Method to compute the degree days, by default ``"integral"``. Possibilities
            are:

            - "integral": integral the degree days above or below the base temperature.
            - "mean": sum the difference between the base temperature and the mean
              temperature.
            - "min_max": sum the difference between the base temperature and the mean
              temperature computed as the mean of the minimum and maximum temperature.

        interseason_mean_temperature : float, optional
            Mean temperature to detect the heating and cooling periods, by default 20.
            Used only:

            - to detect the degree days type automatically. See the
              `detect_degree_days_type` method.
            - to estimate the base temperature. See the `calibrate_base_temperature`
              method.

        base_logger_name : str, optional
            Name of the logger. By default, it is the class name. All following
            following instances receive a unique identifier, based on the first
            one, with the pattern:

            - ``<base_logger_name>``
            - ``<base_logger_name>_1``
            - ...

        min_logger_level_stdout : str or int, optional
            The minimum severity level for messages to be printed to standard output.
            Messages below this level will not be displayed. The default is
            ``"ERROR"``, meaning only errors and critical messages will
            be printed. Acceptable values can be specified as strings (e.g.,
            "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL") or as corresponding
            integer constants from the `logging` module. If you want to have details
            on what's going on in the code, set this to ``"INFO"`` or ``"WARNING"``.

        """
        degree_days_base_temperature = {} or degree_days_base_temperature
        self._categories = categories
        super().__init__(
            energy_data=energy_data,
            temperature_data=temperature_data,
            frequency=frequency,
            degree_days_type=degree_days_type,
            degree_days_base_temperature=degree_days_base_temperature,
            degree_days_computation_method=degree_days_computation_method,
            interseason_mean_temperature=interseason_mean_temperature,
            base_logger_name=base_logger_name,
            min_logger_level_stdout=min_logger_level_stdout,
        )

    @cached_property
    def resampled_categories(
        self,
    ) -> pd.Series:
        """Resample categories at the specified frequency.

        This method resamples the categorical data (self.categories) at the given
        frequency (self.frequency) and computes the most common category within each
        resampled period. If a period contains no data, it returns None for that
        period. The method uses an efficient aggregation approach by leveraging `agg()`
        and `value_counts()` to find the most frequent category.

        The result is a `pd.Series` with the same frequency as the resampling,
        containing the most common category for each resampled time window.

        This property is cached to avoid recomputing it multiple times, improving
        performance in case of repeated access.

        Returns
        -------
        pd.Series
            A Series where each value corresponds to the most common category in the
            resampled period, indexed by the resampled time index.

        Example
        -------
        >>> self.frequency = '1D'  # Daily resampling
        >>> self.categories = pd.Series(['A', 'B', 'A', 'A', 'B'],
        index=pd.date_range('2023-01-01', periods=5))
        >>> resampled = self.resampled_categories
        >>> print(resampled)
        2023-01-01    A
        2023-01-02    A
        2023-01-03    A
        2023-01-04    B
        Freq: D, dtype: object

        .. note::
            In case there are multiple categories for one resampled period, the
            category assigned to the resampled period is the most common one.

        """
        categories: pd.Series = self.categories
        resampler: Resampler = categories.resample(self.frequency)
        resampled_categories: pd.Series = resampler.agg(
            lambda x: x.value_counts().idxmax() if not x.empty else None,
        )
        return resampled_categories.rename(self.categories_name)

    @property
    def categories(
        self,
    ) -> pd.Series:
        """The categories of the periods."""
        return self._categories

    @categories.setter
    def categories(
        self,
        value: pd.Series,
    ) -> None:
        """Set the categories attribute.

        Parameters
        ----------
        value : pd.Series
            A pandas Series to assign as the categories.

        """
        self._categories = value

    @cached_property
    def resampled_energy_temperature_category(
        self,
    ) -> pd.DataFrame:
        """The resampled energy, temperature and category data.

        The DataFrame contains the resampled energy and temperature data.
        Periods with missing values are removed.
        """
        return pd.concat(
            [
                self.resampled_energy,
                self.resampled_temperature,
                self.resampled_categories,
            ],
            axis=1,
        ).dropna(how="any", axis=0)

    def _aggregate_data(
        self,
        degree_days_base_temperature: dict,
    ) -> None:
        """Compute the degree days and aggregate the data.

        Store the aggregated data in the `aggregated_data` property.
        """
        degree_days = self._calculate_degree_days(degree_days_base_temperature)
        self.aggregated_data = pd.concat(
            [
                self.resampled_energy_temperature_category,
                degree_days,
            ],
            axis=1,
        )

    def _detect_degree_days_type(
        self,
        significance_level: float = 0.05,
    ) -> None:
        """Detect the degree days type (heating, cooling, or both).

        Detection of the degree-days types is based on correlations between
        energy consumption and temperature, across different categories.

        This method automatically detects the type of degree days (heating, cooling, or
        both) if the `degree_days_type` is set to "auto". It manages potential
        Simpson's paradox by calculating the Spearman correlation for each category in
        the data, rather than relying on aggregated data. The detection is performed by
        evaluating the correlation between energy consumption and temperature in each
        resampled category.

        If the Spearman correlation between energy consumption and temperature is
        significant for any category:

        - A negative correlation for periods below the interseason mean temperature
          indicates "heating" degree days.
        - A positive correlation for periods above the interseason mean temperature
          indicates "cooling" degree days.
        - If both correlations are significant, the system is assumed to have both
          heating and cooling sensitivity.

        Parameters
        ----------
        significance_level : float, optional
            The threshold for the p-value to consider the correlation as statistically
            significant. The default is 0.05 (5% significance level).

        Returns
        -------
        None
            The method updates the `degree_days_type` attribute of the class, setting
            it to either "heating", "cooling", or "both" based on the correlation
            results. If no significant correlation is found, a `ValueError` is raised.

        Raises
        ------
        ValueError
            If no significant correlation can be detected for either heating or cooling
            degree days, the method raises a `ValueError` and suggests manual
            specification of the `degree_days_type`.

        Notes
        -----
        - **Spearman Correlation**: This non-parametric correlation test is used to
          determine monotonic relationships between energy consumption and temperature.
          It is robust to non-linear relationships and does not assume normality.
        - **Simpson's Paradox**: By evaluating correlations within each category, the
          method avoids potential misleading conclusions that might arise from
          aggregated data. Simpson's Paradox occurs when trends that appear in
          different groups of data disappear or reverse when the data is combined.

        See Also
        --------
        :py:func:`ThermoSensitivity._detect_degree_days_type` : Base implementation for
        non-categorical data.

        Example
        -------
        >>> self._detect_degree_days_type(significance_level=0.05)
        Detected degree days type: heating

        """
        if self.degree_days_type == "auto":
            heating_sp = 1
            cooling_sp = 1
            for cat in list(self.resampled_categories.unique()):
                log_message = f"{cat=}"
                self.logger.info(log_message)
                heating_mask = (
                    self.resampled_energy_temperature_category[self.temperature_name]
                    < self.interseason_mean_temperature
                )
                cat_mask = (
                    self.resampled_energy_temperature_category[self.categories_name]
                    == cat
                )
                tmp_heating_sp = spearmanr(
                    self.resampled_energy_temperature_category.loc[
                        heating_mask & cat_mask,
                        [self.target_name, self.temperature_name],
                    ],
                    alternative="less",
                )
                possib_heat_sp_no_nan = [
                    el for el in [tmp_heating_sp.pvalue, heating_sp] if ~np.isnan(el)
                ]
                heating_sp = min(possib_heat_sp_no_nan)
                cooling_mask = (
                    self.resampled_temperature > self.interseason_mean_temperature
                )
                tmp_cooling_sp = spearmanr(
                    self.resampled_energy_temperature_category.loc[
                        cooling_mask & cat_mask,
                        [self.target_name, self.temperature_name],
                    ],
                    alternative="greater",
                )
                log_message = f"{tmp_cooling_sp}"
                self.logger.info(log_message)
                possib_cool_sp_no_nan = [
                    el for el in [tmp_cooling_sp.pvalue, cooling_sp] if ~np.isnan(el)
                ]
                cooling_sp = min(possib_cool_sp_no_nan)
            if heating_sp < significance_level and cooling_sp < significance_level:
                self.degree_days_type = "both"
            elif heating_sp < significance_level:
                self.degree_days_type = "heating"
            elif cooling_sp < significance_level:
                self.degree_days_type = "cooling"
            else:
                err = (
                    "Cannot detect the degree days type. Please specify it manually. "
                    f"For info, {cooling_sp=}, {heating_sp=}"
                )
                raise ValueError(err)

    def _fit_thermosensitivity(
        self,
    ) -> None:
        """Fit the thermosensitivity model with interactions for categories.

        This method performs the following steps:

        1. Drops rows with missing values from the aggregated data.
        2. Creates a regression model where the dependent variable is energy consumption
          and the independent variables are the degree days (heating, cooling, or both)
          and category interactions.
        3. Encodes the category labels using one-hot encoding.
        4. Constructs interaction terms between the degree days and the one-hot encoded
          categories.
        5. Fits the Ordinary Least Squares (OLS) model to the data with the degree days
          and interaction terms.

        The fitted model is stored in the ``_model`` attribute.

        """
        data = self.aggregated_data.dropna(how="any", axis=0)
        y = data[self.target_name].copy()
        x = data[self.predictors].copy()
        x["Intercept"] = 1  # add constant
        one_hot_encoding = pd.get_dummies(data["category"], dtype=int)
        # create interaction terms
        interactions = None
        for col in x.columns:
            for cat in one_hot_encoding.columns:
                tmp_int = x[col] * one_hot_encoding[cat]
                tmp_int.name = f"{col}:{cat}"
                if interactions is None:
                    interactions = tmp_int
                else:
                    interactions = pd.concat([interactions, tmp_int], axis=1)
        self._model = OLS(y, interactions).fit()



"""Module for computing threshold profiles based on relative deviations.

This module provides tools to calculate profiles that deviate from the mean profile
by a user-defined relative offset. The profiles are designed to scale proportionally
with the baseline consumption pattern derived from historical data.
"""

import pandas as pd

from energy_analysis_toolbox.timeseries.profiles.mean_profile import MeanProfile


class RelativeThreshold(MeanProfile):
    """A class which implements a relative deviation from the mean profile."""

    def __init__(
        self,
        offset_relative: float = 0.5,
        **kwargs,
    ) -> None:
        """Return a threshold profile.

        The threshold profile is obtained using a user-defined relative variation
        from the mean profile built from history.

        Parameters
        ----------
        offset_rel : float, optional
            Relative difference VS the computed reference to obtain
            the threshold profile. Default is 0.5 (profile is 150%
            of reference)

        """
        self.offset_rel = offset_relative
        super().__init__(**kwargs)

    def compute(
        self,
        history: pd.Series,
        time: pd.Timestamp,
        **kwargs,
    ) -> pd.Series:
        """Return a threshold profile.

        The threshold profile is obtained using a user-defined relative variation
        from the mean profile built from history.

        Parameters
        ----------
        history : pd.Series
            Consumption history used to computed the reference
            profile.
        time : pd.Timestamp
            The time at which the profile is of interest. Only
            the information about the date is used in the passed
            timestamp.

        Returns
        -------
        profile : pd.Series
            Profile threshold with same resolution as the history data.

        Notes
        -----
        The profile threshold is obtained as ``(1 + tshd) *`` the mean
        profile obtained from history.

        """
        reference = super().compute(history, time, **kwargs)
        offset = self.offset_rel * reference
        offset.index = reference.index
        return reference + offset



import pandas as pd

from ..power.overconsumption.transform import merge_by_proximity
from .test_power_overconsumption_select import example_intervals


def middle_merged():
    intervals = example_intervals()
    merged = pd.Series(
        [
            intervals["start"].iloc[1],
            intervals["end"].iloc[2],
            (intervals["end"].iloc[2] - intervals["start"].iloc[1]).total_seconds(),
            intervals["energy"].iloc[1:3].sum(),
        ],
        index=intervals.columns,
    )
    merged_intervals = pd.DataFrame.from_dict(
        {
            0: intervals.iloc[0, :],
            1: merged,
            2: intervals.iloc[3, :],
        },
        orient="index",
    )
    return merged_intervals


def test_merge_on_empty():
    intervals = example_intervals().iloc[:0, :]
    assert merge_by_proximity(intervals).empty


def test_merge_no_hit():
    """Check case when the threshold is lower than any duration between overshoots"""
    intervals = example_intervals()
    pd.testing.assert_frame_equal(
        intervals,
        merge_by_proximity(intervals, min_interval=60),
        check_like=True,
        check_dtype=False,
    )


def test_merge_all():
    """Check case when the threshold is big enough to merge all overconsumption."""
    intervals = example_intervals()
    merged = merge_by_proximity(intervals, min_interval=3600 * 24)
    expected = pd.DataFrame.from_dict(
        {
            "start": [intervals["start"].iloc[0]],
            "end": [intervals["end"].iloc[-1]],
            "duration": [
                (intervals["end"].iloc[-1] - intervals["start"].iloc[0]).total_seconds(),
            ],
            "energy": [intervals["energy"].sum()],
        },
    )
    pd.testing.assert_frame_equal(merged, expected, check_dtype=False)


def test_merge_nominal():
    """Check a normal case with a few hits : middle interval merged"""
    intervals = example_intervals()
    merged = merge_by_proximity(intervals, min_interval=600)
    pd.testing.assert_frame_equal(merged, middle_merged(), check_dtype=False)


def test_merge_touching_interval_zero_tshd():
    """Check robustness in limit case when overconsumption are adjacent."""
    intervals = example_intervals()
    intervals.loc[1, "end"] = intervals.loc[2, "start"]
    merged = merge_by_proximity(intervals, min_interval=0)
    pd.testing.assert_frame_equal(merged, middle_merged(), check_dtype=False)
